from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, BinaryIO

from dissect.ntfs.c_ntfs import (
    ATTRIBUTE_TYPE_CODE,
    DEFAULT_CLUSTER_SIZE,
    DEFAULT_INDEX_SIZE,
    DEFAULT_RECORD_SIZE,
    DEFAULT_SECTOR_SIZE,
    FILE_NUMBER_VOLUME,
    NTFS_SIGNATURE,
    c_ntfs,
)
from dissect.ntfs.exceptions import Error, FileNotFoundError, VolumeNotAvailableError
from dissect.ntfs.mft import Mft, MftRecord
from dissect.ntfs.secure import Secure
from dissect.ntfs.usnjrnl import UsnJrnl
from dissect.ntfs.util import bsf

if TYPE_CHECKING:
    from collections.abc import Iterator


class NTFS:
    """Implementation for Microsoft NTFS.

    This implementation supports parsing NTFS from either a full NTFS volume or from separate files.
    If you have a file-like object of an NTFS volume, simply pass it as the fh argument. If you have separate
    file-like objects for things like ``$BOOT`` or ``$MFT``, pass those as the boot and mft arguments.
    The separate arguments take precedence over parsing from the volume file-like object.

    Args:
        fh: A file-like object for the volume to use for parsing NTFS. This is where "data on disk" is read from.
        boot: A file-like object for the ``$BOOT`` file.
        mft: A file-like object for the ``$MFT`` file.
        usnjrnl: A file-like object for the ``$Extend/$Usnjrnl:$J`` file.
        sds: A file-like object for the ``$Secure:$SDS`` file.
    """

    def __init__(
        self,
        fh: BinaryIO | None = None,
        boot: BinaryIO | None = None,
        mft: BinaryIO | None = None,
        usnjrnl: BinaryIO | None = None,
        sds: BinaryIO | None = None,
    ):
        self.fh = fh

        boot_fh = boot or fh
        if boot_fh:
            boot_fh.seek(0)
            self.boot_sector = c_ntfs.BOOT_SECTOR(boot_fh)

            if self.boot_sector.Oem != NTFS_SIGNATURE:
                raise Error(f"Invalid NTFS magic: {self.boot_sector.Oem}")

            self.sector_size = self.boot_sector.Bpb.BytesPerSector
            if self.boot_sector.Bpb.SectorsPerCluster & 0xFF > 0x80:
                sectors_per_cluster = 1 << (-self.boot_sector.Bpb.SectorsPerCluster)
            else:
                sectors_per_cluster = self.boot_sector.Bpb.SectorsPerCluster & 0xFF
            self.cluster_size = sectors_per_cluster * self.sector_size

            if self.boot_sector.ClustersPerFileRecordSegment < 0:
                self._record_size = 1 << (-self.boot_sector.ClustersPerFileRecordSegment)
            else:
                self._record_size = self.boot_sector.ClustersPerFileRecordSegment * self.cluster_size

            if self.boot_sector.ClustersPerIndexBuffer < 0:
                self._index_size = 2 << (-self.boot_sector.ClustersPerIndexBuffer)
            else:
                self._index_size = self.boot_sector.ClustersPerIndexBuffer * self.cluster_size
        else:
            # Provide some defaults if we don't have a volume or $BOOT file
            self.boot_sector = None
            self.sector_size = DEFAULT_SECTOR_SIZE
            self.cluster_size = DEFAULT_CLUSTER_SIZE
            self._record_size = DEFAULT_RECORD_SIZE
            self._index_size = DEFAULT_INDEX_SIZE

        self.sector_size_shift = bsf(self.sector_size)
        self.cluster_size_shift = bsf(self.cluster_size)

        self.mft = None
        self.secure = None
        self.usnjrnl = None

        if mft:
            self.mft = Mft(mft, ntfs=self)
        elif self.fh:
            # Small preface for what's happening here:
            # - MFT records have one or more $DATA attributes to describe where data is stored on disk
            # - When data is extremely fragmented, you need more than one $DATA attribute
            # - When you run out of space in the MFT record itself, attributes are stored in other MFT records
            #   and referenced by an $ATTRIBUTE_LIST in the main MFT record
            # - To be able to look up these other MFT records, you need to be able to find them in the $MFT file
            #
            # This is all fine, until you realize that the $MFT file itself can be extremely fragmented,
            # creating a chicken-egg problem. We need the MFT in order to open the $MFT file.

            # First, parse just the MFT record
            mft_offset = self.boot_sector.MftStartLcn * self.cluster_size
            mft_record = MftRecord.from_fh(fh, mft_offset, ntfs=self)

            # Second, we open a temporary MFT with the resident $DATA attributes we parsed from the MFT record
            # This is because if there's no MFT loaded yet (mft attribute on this class) we don't load attribute lists
            self.mft = Mft(mft_record.open(), ntfs=self)

            # With this initial MFT object we can properly resolve attribute list entries, since the next MFT record
            # containing the next $DATA attribute is guaranteed to be within the $DATA runs that we already parsed
            if ATTRIBUTE_TYPE_CODE.ATTRIBUTE_LIST in mft_record.attributes:
                for datarun in _get_dataruns_from_attribute_list(mft_record):
                    # Update the RunlistStream of the MFT with new runs as we find them as the next
                    # attribute list entry may point to an MFT record that is contained in the previous
                    # attribute list entry's $DATA
                    self.mft.fh.runlist += datarun

        if sds:
            self.secure = Secure(sds=sds)
        elif self.mft:
            try:
                self.secure = Secure(self.mft.get("$Secure"))
            except (FileNotFoundError, VolumeNotAvailableError):
                pass

        if usnjrnl:
            self.usnjrnl = UsnJrnl(usnjrnl, ntfs=self)
        elif self.mft:
            try:
                self.usnjrnl = UsnJrnl(self.mft.get("$Extend\\$Usnjrnl").open("$J"), ntfs=self)
            except (FileNotFoundError, NotADirectoryError, VolumeNotAvailableError):
                pass

    @cached_property
    def serial(self) -> int | None:
        return self.boot_sector.SerialNumber if self.boot_sector else None

    @cached_property
    def volume_name(self) -> str | None:
        if not self.mft:
            return None

        try:
            volume_file = self.mft.get(FILE_NUMBER_VOLUME)
            volume_name = volume_file.attributes[ATTRIBUTE_TYPE_CODE.VOLUME_NAME].data()
            return volume_name.decode("utf-16-le")
        except (AttributeError, FileNotFoundError):
            return None


def _get_dataruns_from_attribute_list(record: MftRecord) -> Iterator[list[tuple[int, int]]]:
    for attr in record.attributes[ATTRIBUTE_TYPE_CODE.ATTRIBUTE_LIST].attributes():
        if attr.type == ATTRIBUTE_TYPE_CODE.DATA:
            yield attr.dataruns()



from __future__ import annotations

from dissect.cstruct import cstruct

ntfs_def = """
/* ================ Generic stuff ================ */

flag FILE_ATTRIBUTE : DWORD {
    READONLY                = 0x00000001,
    HIDDEN                  = 0x00000002,
    SYSTEM                  = 0x00000004,
    DIRECTORY               = 0x00000010,
    ARCHIVE                 = 0x00000020,
    DEVICE                  = 0x00000040,
    NORMAL                  = 0x00000080,
    TEMPORARY               = 0x00000100,
    SPARSE_FILE             = 0x00000200,
    REPARSE_POINT           = 0x00000400,
    COMPRESSED              = 0x00000800,
    OFFLINE                 = 0x00001000,
    NOT_CONTENT_INDEXED     = 0x00002000,
    ENCRYPTED               = 0x00004000,
    INTEGRITY_STREAM        = 0x00008000,
    VIRTUAL                 = 0x00010000,
    NO_SCRUB_DATA           = 0x00020000,
    RECALL_ON_OPEN          = 0x00040000,
    PINNED                  = 0x00080000,
    UNPINNED                = 0x00100000,
    RECALL_ON_DATA_ACCESS   = 0x00400000,
};

/* ================ Volume headers ================ */

typedef struct _BIOS_PARAMETER_BLOCK {
    USHORT      BytesPerSector;
    INT8        SectorsPerCluster;
    USHORT      ReservedSectors;
    UCHAR       Fats;
    USHORT      RootEntries;
    USHORT      Sectors;
    UCHAR       Media;
    USHORT      SectorsPerFat;
    USHORT      SectorsPerTrack;
    USHORT      Heads;
    ULONG       HiddenSectors;
    ULONG       LargeSectors;
} BIOS_PARAMETER_BLOCK;

typedef struct _BOOT_SECTOR {
    CHAR        Jump[3];
    CHAR        Oem[8];
    BIOS_PARAMETER_BLOCK    Bpb;
    CHAR        Unused0[4];
    ULONG64     NumberSectors;
    ULONG64     MftStartLcn;
    ULONG64     Mft2StartLcn;
    INT8        ClustersPerFileRecordSegment;
    CHAR        Reserved0[3];
    INT8        ClustersPerIndexBuffer;
    CHAR        Reserved1[3];
    ULONG64     SerialNumber;
    ULONG       Checksum;
    CHAR        BootStrap[0x200-0x054];
} BOOT_SECTOR;

/* ================ MFT stuff ================ */

typedef struct _MFT_SEGMENT_REFERENCE {
    ULONG       SegmentNumberLowPart;
    USHORT      SegmentNumberHighPart;
    USHORT      SequenceNumber;
} MFT_SEGMENT_REFERENCE;

typedef MFT_SEGMENT_REFERENCE FILE_REFERENCE;

typedef struct _MULTI_SECTOR_HEADER {
    CHAR        Signature[4];
    USHORT      UpdateSequenceArrayOffset;
    USHORT      UpdateSequenceArraySize;
} MULTI_SECTOR_HEADER;

typedef struct _FILE_RECORD_SEGMENT_HEADER {
    MULTI_SECTOR_HEADER MultiSectorHeader;
    ULONG64     Lsn;
    USHORT      SequenceNumber;
    USHORT      ReferenceCount;
    USHORT      FirstAttributeOffset;
    USHORT      Flags;
    ULONG       BytesInUse;
    ULONG       BytesAllocated;
    FILE_REFERENCE  BaseFileRecordSegment;
    USHORT      NextAttributeInstance;
} FILE_RECORD_SEGMENT_HEADER;

/* ================ Attributes ================ */

enum ATTRIBUTE_TYPE_CODE : ULONG {
    UNUSED                  = 0x00,
    STANDARD_INFORMATION    = 0x10,
    ATTRIBUTE_LIST          = 0x20,
    FILE_NAME               = 0x30,
    OBJECT_ID               = 0x40,
    SECURITY_DESCRIPTOR     = 0x50,
    VOLUME_NAME             = 0x60,
    VOLUME_INFORMATION      = 0x70,
    DATA                    = 0x80,
    INDEX_ROOT              = 0x90,
    INDEX_ALLOCATION        = 0xA0,
    BITMAP                  = 0xB0,
    REPARSE_POINT           = 0xC0,
    EA_INFORMATION          = 0xD0,
    EA                      = 0xE0,
    PROPERTY_SET            = 0xF0,
    LOGGED_UTILITY_STREAM   = 0x100,
    END                     = 0xFFFFFFFF,
};

typedef struct _ATTRIBUTE_RECORD_HEADER {
    ATTRIBUTE_TYPE_CODE TypeCode;
    ULONG       RecordLength;
    UCHAR       FormCode;
    UCHAR       NameLength;
    USHORT      NameOffset;
    USHORT      Flags;
    USHORT      Instance;
    union {
        struct {
            ULONG   ValueLength;
            USHORT  ValueOffset;
            UCHAR   Flags;
            UCHAR   Reserved;
        } Resident;
        struct {
            ULONG64     LowestVcn;
            ULONG64     HighestVcn;
            USHORT      MappingPairsOffset;
            UCHAR       CompressionUnit;
            UCHAR       Reserved[5];
            LONGLONG    AllocatedLength;
            LONGLONG    FileSize;
            LONGLONG    ValidDataLength;
            LONGLONG    TotalAllocated;
        } Nonresident;
    } Form;
} ATTRIBUTE_RECORD_HEADER;

typedef struct _STANDARD_INFORMATION {
    LONGLONG    CreationTime;
    LONGLONG    LastModificationTime;
    LONGLONG    LastChangeTime;
    LONGLONG    LastAccessTime;
    ULONG       FileAttributes;
    ULONG       MaximumVersions;
    ULONG       VersionNumber;
    ULONG       ClassId;
    ULONG       OwnerId;
    ULONG       SecurityId;
    ULONGLONG   QuotaCharged;
    ULONGLONG   Usn;
} STANDARD_INFORMATION_EX;

typedef struct _ATTRIBUTE_LIST_ENTRY {
    ATTRIBUTE_TYPE_CODE AttributeTypeCode;
    USHORT      RecordLength;
    UCHAR       AttributeNameLength;
    UCHAR       AttributeNameOffset;
    ULONG64     LowestVcn;
    MFT_SEGMENT_REFERENCE   SegmentReference;
    USHORT      Reserved;
    WCHAR       AttributeName[AttributeNameLength];
} ATTRIBUTE_LIST_ENTRY;

typedef struct _FILE_NAME {
    FILE_REFERENCE  ParentDirectory;
    LONGLONG    CreationTime;
    LONGLONG    LastModificationTime;
    LONGLONG    LastChangeTime;
    LONGLONG    LastAccessTime;
    LONGLONG    AllocatedLength;
    LONGLONG    FileSize;
    ULONG       FileAttributes;
    union {
        struct {
            USHORT  EaSize;
            USHORT  _;
        };
        ULONG   ReparsePointTag;
    };
    UCHAR       FileNameLength;
    UCHAR       Flags;
    WCHAR       FileName[FileNameLength];
} FILE_NAME;

enum IO_REPARSE_TAG : ULONG {
    RESERVED_ZERO           = 0x00000000,
    RESERVED_ONE            = 0x00000001,
    RESERVED_TWO            = 0x00000002,
    MOUNT_POINT             = 0xA0000003,
    HSM                     = 0xC0000004,
    DRIVE_EXTENDER          = 0x80000005,
    HSM2                    = 0x80000006,
    SIS                     = 0x80000007,
    WIM                     = 0x80000008,
    CSV                     = 0x80000009,
    DFS                     = 0x8000000A,
    FILTER_MANAGER          = 0x8000000B,
    SYMLINK                 = 0xA000000C,
    IIS_CACHE               = 0xA0000010,
    DFSR                    = 0x80000012,
    DEDUP                   = 0x80000013,
    APPXSTRM                = 0xC0000014,
    NFS                     = 0x80000014,
    FILE_PLACEHOLDER        = 0x80000015,
    DFM                     = 0x80000016,
    WOF                     = 0x80000017,
    WCI                     = 0x80000018,
    WCI_1                   = 0x90001018,
    GLOBAL_REPARSE          = 0xA0000019,
    CLOUD                   = 0x9000001A,
    CLOUD_1                 = 0x9000101A,
    CLOUD_2                 = 0x9000201A,
    CLOUD_3                 = 0x9000301A,
    CLOUD_4                 = 0x9000401A,
    CLOUD_5                 = 0x9000501A,
    CLOUD_6                 = 0x9000601A,
    CLOUD_7                 = 0x9000701A,
    CLOUD_8                 = 0x9000801A,
    CLOUD_9                 = 0x9000901A,
    CLOUD_A                 = 0x9000A01A,
    CLOUD_B                 = 0x9000B01A,
    CLOUD_C                 = 0x9000C01A,
    CLOUD_D                 = 0x9000D01A,
    CLOUD_E                 = 0x9000E01A,
    CLOUD_F                 = 0x9000F01A,
    APPEXECLINK             = 0x8000001B,
    PROJFS                  = 0x9000001C,
    LX_SYMLINK              = 0xA000001D,
    STORAGE_SYNC            = 0x8000001E,
    WCI_TOMBSTONE           = 0xA000001F,
    UNHANDLED               = 0x80000020,
    ONEDRIVE                = 0x80000021,
    PROJFS_TOMBSTONE        = 0xA0000022,
    AF_UNIX                 = 0x80000023,
    LX_FIFO                 = 0x80000024,
    LX_CHR                  = 0x80000025,
    LX_BLK                  = 0x80000026,
    WCI_LINK                = 0xA0000027,
    WCI_LINK_1              = 0xA0001027,
};

typedef struct _REPARSE_DATA_BUFFER {
    IO_REPARSE_TAG  ReparseTag;
    USHORT      ReparseDataLength;
    USHORT      Reserved;
} REPARSE_DATA_BUFFER;

enum SYMLINK_FLAG : ULONG {
    ABSOLUTE                = 0x00000000,
    RELATIVE                = 0x00000001,
};

typedef struct _SYMBOLIC_LINK_REPARSE_BUFFER {
    USHORT  SubstituteNameOffset;
    USHORT  SubstituteNameLength;
    USHORT  PrintNameOffset;
    USHORT  PrintNameLength;
    SYMLINK_FLAG    Flags;
} SYMBOLIC_LINK_REPARSE_BUFFER;

typedef struct _MOUNT_POINT_REPARSE_BUFFER {
    USHORT  SubstituteNameOffset;
    USHORT  SubstituteNameLength;
    USHORT  PrintNameOffset;
    USHORT  PrintNameLength;
} _MOUNT_POINT_REPARSE_BUFFER;

typedef struct _CLOUD_FILTER_REPARSE_BUFFER {
    // ULONG   Unknown_1;
    // ULONG   Unknown_2;
    CHAR    Guid[16];
    USHORT  NameLength;
    // WCHAR    Name[NameLength];
} _CLOUD_FILTER_REPARSE_BUFFER;

/* ================ Index ================ */

enum COLLATION : ULONG {
    BINARY                  = 0,
    FILE_NAME               = 1,
    UNICODE_STRING          = 2,
    NUMBER_RULES            = 3,
    NTOFS_ULONG             = 16,
    NTOFS_SID               = 17,
    NTOFS_SECURITY_HASH     = 18,
    NTOFS_ULONGS            = 19,
};

typedef COLLATION COLLATION_RULE;

typedef struct _INDEX_HEADER {
    ULONG       FirstEntryOffset;
    ULONG       TotalSizeOfEntries;
    ULONG       AllocatedSize;
    UCHAR       Flags;
    UCHAR       Reserved[3];
} INDEX_HEADER;

typedef struct _INDEX_ROOT {
    ATTRIBUTE_TYPE_CODE AttributeType;
    COLLATION_RULE  CollationRule;
    ULONG       BytesPerIndexBuffer;
    UCHAR       ClustersPerIndexBuffer;
    UCHAR       Reserved[3];
    INDEX_HEADER    IndexHeader;
} INDEX_ROOT;

typedef struct _INDEX_ALLOCATION_BUFFER {
    MULTI_SECTOR_HEADER     MultiSectorHeader;
    ULONG64     Lsn;
    ULONG64     Vcn;
    INDEX_HEADER    IndexHeader;
} INDEX_ALLOCATION_BUFFER;

typedef struct _INDEX_ENTRY {
    union {
        FILE_REFERENCE  FileReference;
        struct {
            USHORT      DataOffset;
            USHORT      DataLength;
            ULONG       _;
        };
    };
    USHORT      Length;
    USHORT      KeyLength;
    USHORT      Flags;
    USHORT      Reserved;
} INDEX_ENTRY;

/* ================ Security Descriptors ================ */

flag SECURITY_DESCRIPTOR_CONTROL : WORD {
    SE_OWNER_DEFAULTED          = 0x0001,
    SE_GROUP_DEFAULTED          = 0x0002,
    SE_DACL_PRESENT             = 0x0004,
    SE_DACL_DEFAULTED           = 0x0008,
    SE_SACL_PRESENT             = 0x0010,
    SE_SACL_DEFAULTED           = 0x0020,
    SE_DACL_AUTO_INHERIT_REQ    = 0x0100,
    SE_SACL_AUTO_INHERIT_REQ    = 0x0200,
    SE_DACL_AUTO_INHERITED      = 0x0400,
    SE_SACL_AUTO_INHERITED      = 0x0800,
    SE_DACL_PROTECTED           = 0x1000,
    SE_SACL_PROTECTED           = 0x2000,
    SE_RM_CONTROL_VALID         = 0x4000,
    SE_SELF_RELATIVE            = 0x8000,
};

flag ACCESS_MASK : DWORD {
    FILE_READ_DATA              = 0x00000001,
    FILE_LIST_DIRECTORY         = 0x00000001,
    FILE_WRITE_DATA             = 0x00000002,
    FILE_ADD_FILE               = 0x00000002,
    FILE_APPEND_DATA            = 0x00000004,
    FILE_ADD_SUBDIRECTORY       = 0x00000004,
    FILE_READ_EA                = 0x00000008,
    FILE_WRITE_EA               = 0x00000010,
    FILE_EXECUTE                = 0x00000020,
    FILE_TRAVERSE               = 0x00000020,
    FILE_DELETE_CHILD           = 0x00000040,
    FILE_READ_ATTRIBUTES        = 0x00000080,
    FILE_WRITE_ATTRIBUTES       = 0x00000100,
    DELETE                      = 0x00010000,
    READ_CONTROL                = 0x00020000,
    WRITE_DAC                   = 0x00040000,
    WRITE_OWNER                 = 0x00080000,
    SYNCHRONIZE                 = 0x00100000,
    STANDARD_RIGHTS_READ        = 0x00020000,
    STANDARD_RIGHTS_WRITE       = 0x00020000,
    STANDARD_RIGHTS_EXECUTE     = 0x00020000,
    STANDARD_RIGHTS_REQUIRED    = 0x000f0000,
    STANDARD_RIGHTS_ALL         = 0x001f0000,
    ACCESS_SYSTEM_SECURITY      = 0x01000000,
    MAXIMUM_ALLOWED             = 0x02000000,
    GENERIC_ALL                 = 0x10000000,
    GENERIC_EXECUTE             = 0x20000000,
    GENERIC_WRITE               = 0x40000000,
    GENERIC_READ                = 0x80000000,
};

enum ACE_TYPE : BYTE {
    ACCESS_ALLOWED                  = 0x00,
    ACCESS_DENIED                   = 0x01,
    SYSTEM_AUDIT                    = 0x02,
    SYSTEM_ALARM                    = 0x03,
    ACCESS_ALLOWED_COMPOUND         = 0x04,
    ACCESS_ALLOWED_OBJECT           = 0x05,
    ACCESS_DENIED_OBJECT            = 0x06,
    SYSTEM_AUDIT_OBJECT             = 0x07,
    SYSTEM_ALARM_OBJECT             = 0x08,
    ACCESS_ALLOWED_CALLBACK         = 0x09,
    ACCESS_DENIED_CALLBACK          = 0x0A,
    ACCESS_ALLOWED_CALLBACK_OBJECT  = 0x0B,
    ACCESS_DENIED_CALLBACK_OBJECT   = 0x0C,
    SYSTEM_AUDIT_CALLBACK           = 0x0D,
    SYSTEM_ALARM_CALLBACK           = 0x0E,
    SYSTEM_AUDIT_CALLBACK_OBJECT    = 0x0F,
    SYSTEM_ALARM_CALLBACK_OBJECT    = 0x10,
    SYSTEM_MANDATORY_LABEL          = 0x11,
    SYSTEM_RESOURCE_ATTRIBUTE       = 0x12,
    SYSTEM_SCOPED_POLICY_ID         = 0x13,
    SYSTEM_PROCESS_TRUST_LABEL      = 0x14,
    SYSTEM_ACCESS_FILTER            = 0x15,
};

flag ACE_FLAGS : BYTE {
    OBJECT_INHERIT_ACE          = 0x01,
    CONTAINER_INHERIT_ACE       = 0x02,
    NO_PROPAGATE_INHERIT_ACE    = 0x04,
    INHERIT_ONLY_ACE            = 0x08,
    INHERITED_ACE               = 0x10,
    SUCCESSFUL_ACCESS_ACE_FLAG  = 0x40,
    FAILED_ACCESS_ACE_FLAG      = 0x80,
};

flag ACE_OBJECT_FLAGS : DWORD {
    ACE_OBJECT_TYPE_PRESENT           = 0x01,
    ACE_INHERITED_OBJECT_TYPE_PRESENT = 0x02,
};

enum COMPOUND_ACE_TYPE : USHORT {
    COMPOUND_ACE_IMPERSONATION  = 0x01,
};

typedef struct _ACL {
    BYTE        AclRevision;
    BYTE        Sbz1;
    WORD        AclSize;
    WORD        AceCount;
    WORD        Sbz2;
} ACL;

typedef struct _ACE_HEADER {
    ACE_TYPE    AceType;
    ACE_FLAGS   AceFlags;
    WORD        AceSize;
} ACE_HEADER;

typedef struct _SECURITY_DESCRIPTOR_HEADER {
    ULONG       HashId;
    ULONG       SecurityId;
    ULONG64     Offset;
    ULONG       Length;
} SECURITY_DESCRIPTOR_HEADER;

typedef struct _SECURITY_DESCRIPTOR_RELATIVE {
    BYTE        Revision;
    BYTE        Sbz1;
    SECURITY_DESCRIPTOR_CONTROL Control;
    ULONG       Owner;
    ULONG       Group;
    ULONG       Sacl;
    ULONG       Dacl;
} SECURITY_DESCRIPTOR_RELATIVE;

/* ================ USN Journal ================ */

flag USN_REASON : DWORD {
    DATA_OVERWRITE          = 0x00000001,
    DATA_EXTEND             = 0x00000002,
    DATA_TRUNCATION         = 0x00000004,
    NAMED_DATA_OVERWRITE    = 0x00000010,
    NAMED_DATA_EXTEND       = 0x00000020,
    NAMED_DATA_TRUNCATION   = 0x00000040,
    FILE_CREATE             = 0x00000100,
    FILE_DELETE             = 0x00000200,
    EA_CHANGE               = 0x00000400,
    SECURITY_CHANGE         = 0x00000800,
    RENAME_OLD_NAME         = 0x00001000,
    RENAME_NEW_NAME         = 0x00002000,
    INDEXABLE_CHANGE        = 0x00004000,
    BASIC_INFO_CHANGE       = 0x00008000,
    HARD_LINK_CHANGE        = 0x00010000,
    COMPRESSION_CHANGE      = 0x00020000,
    ENCRYPTION_CHANGE       = 0x00040000,
    OBJECT_ID_CHANGE        = 0x00080000,
    REPARSE_POINT_CHANGE    = 0x00100000,
    STREAM_CHANGE           = 0x00200000,
    TRANSACTED_CHANGE       = 0x00400000,
    INTEGRITY_CHANGE        = 0x00800000,
    CLOSE                   = 0x80000000,
};

flag USN_SOURCE : DWORD {
    NORMAL                  = 0x00000000,
    DATA_MANAGEMENT         = 0x00000001,
    AUXILIARY_DATA          = 0x00000002,
    REPLICATION_MANAGEMENT  = 0x00000004,
    CLIENT_REPLICATION_MANAGEMENT   = 0x00000008,
};

typedef struct _FILE_ID_128 {
    BYTE        Identifier[16];
} FILE_ID_128;

typedef struct {
    DWORD       RecordLength;
    WORD        MajorVersion;
    WORD        MinorVersion;
} USN_RECORD_COMMON_HEADER;

typedef struct {
    DWORD       RecordLength;
    WORD        MajorVersion;
    WORD        MinorVersion;
    MFT_SEGMENT_REFERENCE   FileReferenceNumber;
    MFT_SEGMENT_REFERENCE   ParentFileReferenceNumber;
    ULONG64     Usn;
    ULONG64     TimeStamp;
    USN_REASON  Reason;
    USN_SOURCE  SourceInfo;
    DWORD       SecurityId;
    FILE_ATTRIBUTE  FileAttributes;
    WORD        FileNameLength;
    WORD        FileNameOffset;
} USN_RECORD_V2;

typedef struct {
    DWORD       RecordLength;
    WORD        MajorVersion;
    WORD        MinorVersion;
    FILE_ID_128 FileReferenceNumber;
    FILE_ID_128 ParentFileReferenceNumber;
    ULONG64     Usn;
    ULONG64     TimeStamp;
    USN_REASON  Reason;
    USN_SOURCE  SourceInfo;
    DWORD       SecurityId;
    FILE_ATTRIBUTE  FileAttributes;
    WORD        FileNameLength;
    WORD        FileNameOffset;
} USN_RECORD_V3;

typedef struct {
    LONGLONG    Offset;
    LONGLONG    Length;
} USN_RECORD_EXTENT;

typedef struct {
    DWORD       RecordLength;
    WORD        MajorVersion;
    WORD        MinorVersion;
    FILE_ID_128 FileReferenceNumber;
    FILE_ID_128 ParentFileReferenceNumber;
    ULONG64     Usn;
    USN_REASON  Reason;
    USN_SOURCE  SourceInfo;
    DWORD       RemainingExtents;
    WORD        NumberOfExtents;
    WORD        ExtentSize;
} USN_RECORD_V4;
"""

c_ntfs = cstruct().load(ntfs_def)

# Useful enums and flags
ATTRIBUTE_TYPE_CODE = c_ntfs.ATTRIBUTE_TYPE_CODE
IO_REPARSE_TAG = c_ntfs.IO_REPARSE_TAG
ACCESS_MASK = c_ntfs.ACCESS_MASK
ACE_TYPE = c_ntfs.ACE_TYPE
ACE_OBJECT_FLAGS = c_ntfs.ACE_OBJECT_FLAGS
COLLATION = c_ntfs.COLLATION

# Some useful magic numbers and constants
NTFS_SIGNATURE = b"NTFS    "

SECTOR_SIZE = 512
SECTOR_SHIFT = 9

USN_PAGE_SIZE = 4096

DEFAULT_SECTOR_SIZE = 512
DEFAULT_CLUSTER_SIZE = 4096
DEFAULT_RECORD_SIZE = 1024
DEFAULT_INDEX_SIZE = 4096

# File numbers of various system files
FILE_NUMBER_MFT = 0
FILE_NUMBER_MFTMIRR = 1
FILE_NUMBER_LOGFILE = 2
FILE_NUMBER_VOLUME = 3
FILE_NUMBER_ATTRDEF = 4
FILE_NUMBER_ROOT = 5
FILE_NUMBER_BITMAP = 6
FILE_NUMBER_BOOT = 7
FILE_NUMBER_BADCLUS = 8
FILE_NUMBER_SECURE = 9
FILE_NUMBER_UPCASE = 10
FILE_NUMBER_EXTEND = 11

# File record flags
FILE_RECORD_SEGMENT_IN_USE = 0x0001
FILE_FILE_NAME_INDEX_PRESENT = 0x0002

# Attribute flags
ATTRIBUTE_FLAG_COMPRESSION_MASK = 0x00FF
ATTRIBUTE_FLAG_ENCRYPTED = 0x4000
ATTRIBUTE_FLAG_SPARSE = 0x8000

# Filename flags
FILE_NAME_NTFS = 0x01
FILE_NAME_DOS = 0x02

# Compression flags
COMPRESSION_FORMAT_NONE = 0x0000
COMPRESSION_FORMAT_DEFAULT = 0x0001
COMPRESSION_FORMAT_LZNT1 = 0x0002

# Index and index entry flags
INDEX_NODE = 0x01
INDEX_ENTRY_NODE = 0x01
INDEX_ENTRY_END = 0x02



from __future__ import annotations

import pytest

from dissect.ntfs import exceptions


@pytest.mark.parametrize(
    ("exc", "std"),
    [
        (exceptions.FileNotFoundError, FileNotFoundError),
        (exceptions.IsADirectoryError, IsADirectoryError),
        (exceptions.NotADirectoryError, NotADirectoryError),
    ],
)
def test_filesystem_error_subclass(exc: exceptions.Error, std: Exception) -> None:
    assert issubclass(exc, std)
    assert isinstance(exc(), std)

    with pytest.raises(std):
        raise exc()



from __future__ import annotations

import csv
import gzip
import io
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

import pytest
from dissect.util.stream import MappingStream

if TYPE_CHECKING:
    from collections.abc import Iterator


def absolute_path(filename: str) -> Path:
    return Path(__file__).parent / filename


def open_file_gz(name: str, mode: str = "rb") -> Iterator[BinaryIO]:
    with gzip.GzipFile(absolute_path(name), mode) as f:
        yield f


@pytest.fixture
def ntfs_bin() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/ntfs.bin.gz")


@pytest.fixture
def mft_bin() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/mft.bin.gz")


@pytest.fixture
def ntfs_cloud_bin() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/ntfs-cloud.bin.gz")


@pytest.fixture
def sds_bin() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/sds.bin.gz")


@pytest.fixture
def sds_complex_bin() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/sds_complex.bin.gz")


@pytest.fixture
def boot_2m_bin() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/boot_2m.bin.gz")


@pytest.fixture
def ntfs_fragmented_mft_fh() -> BinaryIO:
    # Test data from https://github.com/msuhanov/ntfs-samples
    # This is from the file ntfs_extremely_fragmented_mft.raw which has, as the name implies, a heavily fragmented MFT
    # The entire file is way too large, so only take just enough data that we actually need to make dissect.ntfs happy
    # We use a MappingStream to stitch everything together at the correct offsets

    stream = MappingStream(align=512)
    with io.TextIOWrapper(gzip.open(absolute_path("_data/ntfs_fragmented_mft.csv.gz"), "r")) as fh:
        for offset, data in csv.reader(fh):
            buf = bytes.fromhex(data)
            stream.add(int(offset), len(buf), io.BytesIO(buf), 0)

    return stream



from __future__ import annotations

import io
from typing import BinaryIO

from dissect.util import lznt1
from dissect.util.stream import RunlistStream


class CompressedRunlistStream(RunlistStream):
    """Specialized :class:`~dissect.util.stream.RunlistStream` for reading NTFS compressed streams.

    Args:
        fh: The source file-like object.
        runlist: The runlist for this stream in block units.
        size: The size of the stream. This can be smaller than the total sum of blocks (to account for slack space).
        compression_unit: The compression unit of for this stream.
    """

    def __init__(
        self, fh: BinaryIO, runlist: list[tuple[int, int]], size: int, cluster_size: int, compression_unit: int
    ):
        # RunlistStream has block_size but we want to make cluster_size available to be more in line with NTFS naming
        self.cluster_size = cluster_size
        self.compression_unit = compression_unit
        self.compression_unit_size = cluster_size << compression_unit

        self._cu_blocks = []

        # First use the compression_unit_size to set the alignment of AlignedStream
        super().__init__(fh, runlist, size, self.compression_unit_size)
        # This also sets the block_size of RunlistStream, so reset that to the correct cluster_size
        self.block_size = cluster_size

    @property
    def runlist(self) -> list[tuple[int, int]]:
        return self._runlist

    @runlist.setter
    def runlist(self, runlist: list[tuple[int, int]]) -> None:
        self._runlist = runlist

        runs = []
        block_num = 0
        current = 2**self.compression_unit

        # Build a list of aligned CU blocks
        for lcn, count in runlist:
            while count > 0:
                use = min(count, current)
                runs.append((lcn, use))

                current -= use
                count -= use
                if lcn is not None:
                    lcn += use

                if current == 0:
                    block_num += 1
                    self._cu_blocks.append(runs)
                    runs = []
                    current = 2**self.compression_unit

    def _read(self, offset: int, length: int) -> bytes:
        # Compressed data is split over compression units (typically 64KB, or 16 clusters)
        # Each compression unit is exactly this size. Compressed data uncompresses to this size
        # Since the AlignedStream is aligned to this size, we just have to take care of reading
        # these compression units and decompressing them
        read_list = []

        # Calculate which CU block we should be reading from
        cu_block = offset // self.compression_unit_size
        cu_offset = cu_block * self.compression_unit_size

        total_remaining = self.size - cu_offset

        while length > 0:
            buf = []
            buf_len = 0
            compressed = False

            for lcn, run_count in self._cu_blocks[cu_block]:
                # If a run is followed by a sparse run within the same CU, the run is compressed
                if lcn is None:
                    compressed = True
                    continue

                run_len = run_count * self.block_size
                self._fh.seek(lcn * self.block_size)
                buf.append(self._fh.read(run_len))
                buf_len += run_len

                if buf_len == total_remaining:
                    break

            if not buf:
                # Completely sparse CU
                read_list.append(b"\x00" * self.compression_unit_size)
            elif compressed:
                # Compressed CU
                try:
                    buf.append(b"\x00" * 64)
                    read_list.append(lznt1.decompress(io.BytesIO(b"".join(buf)))[: self.compression_unit_size])
                except Exception:
                    raise IOError("Decompression failed")
            else:
                # Uncompressed CU
                read_list.append(b"".join(buf))

            length -= self.compression_unit_size
            total_remaining -= self.compression_unit_size
            cu_block += 1

        return b"".join(read_list)



# Generated by cstruct-stubgen
from typing import BinaryIO, Literal, TypeAlias, overload

import dissect.cstruct as __cs__

class _c_ntfs(__cs__.cstruct):
    class FILE_ATTRIBUTE(__cs__.Flag):
        READONLY = ...
        HIDDEN = ...
        SYSTEM = ...
        DIRECTORY = ...
        ARCHIVE = ...
        DEVICE = ...
        NORMAL = ...
        TEMPORARY = ...
        SPARSE_FILE = ...
        REPARSE_POINT = ...
        COMPRESSED = ...
        OFFLINE = ...
        NOT_CONTENT_INDEXED = ...
        ENCRYPTED = ...
        INTEGRITY_STREAM = ...
        VIRTUAL = ...
        NO_SCRUB_DATA = ...
        RECALL_ON_OPEN = ...
        PINNED = ...
        UNPINNED = ...
        RECALL_ON_DATA_ACCESS = ...

    class _BIOS_PARAMETER_BLOCK(__cs__.Structure):
        BytesPerSector: _c_ntfs.uint16
        SectorsPerCluster: _c_ntfs.int8
        ReservedSectors: _c_ntfs.uint16
        Fats: _c_ntfs.uint8
        RootEntries: _c_ntfs.uint16
        Sectors: _c_ntfs.uint16
        Media: _c_ntfs.uint8
        SectorsPerFat: _c_ntfs.uint16
        SectorsPerTrack: _c_ntfs.uint16
        Heads: _c_ntfs.uint16
        HiddenSectors: _c_ntfs.uint32
        LargeSectors: _c_ntfs.uint32
        @overload
        def __init__(
            self,
            BytesPerSector: _c_ntfs.uint16 | None = ...,
            SectorsPerCluster: _c_ntfs.int8 | None = ...,
            ReservedSectors: _c_ntfs.uint16 | None = ...,
            Fats: _c_ntfs.uint8 | None = ...,
            RootEntries: _c_ntfs.uint16 | None = ...,
            Sectors: _c_ntfs.uint16 | None = ...,
            Media: _c_ntfs.uint8 | None = ...,
            SectorsPerFat: _c_ntfs.uint16 | None = ...,
            SectorsPerTrack: _c_ntfs.uint16 | None = ...,
            Heads: _c_ntfs.uint16 | None = ...,
            HiddenSectors: _c_ntfs.uint32 | None = ...,
            LargeSectors: _c_ntfs.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    BIOS_PARAMETER_BLOCK: TypeAlias = _BIOS_PARAMETER_BLOCK
    class _BOOT_SECTOR(__cs__.Structure):
        Jump: __cs__.CharArray
        Oem: __cs__.CharArray
        Bpb: _c_ntfs._BIOS_PARAMETER_BLOCK
        Unused0: __cs__.CharArray
        NumberSectors: _c_ntfs.uint64
        MftStartLcn: _c_ntfs.uint64
        Mft2StartLcn: _c_ntfs.uint64
        ClustersPerFileRecordSegment: _c_ntfs.int8
        Reserved0: __cs__.CharArray
        ClustersPerIndexBuffer: _c_ntfs.int8
        Reserved1: __cs__.CharArray
        SerialNumber: _c_ntfs.uint64
        Checksum: _c_ntfs.uint32
        BootStrap: __cs__.CharArray
        @overload
        def __init__(
            self,
            Jump: __cs__.CharArray | None = ...,
            Oem: __cs__.CharArray | None = ...,
            Bpb: _c_ntfs._BIOS_PARAMETER_BLOCK | None = ...,
            Unused0: __cs__.CharArray | None = ...,
            NumberSectors: _c_ntfs.uint64 | None = ...,
            MftStartLcn: _c_ntfs.uint64 | None = ...,
            Mft2StartLcn: _c_ntfs.uint64 | None = ...,
            ClustersPerFileRecordSegment: _c_ntfs.int8 | None = ...,
            Reserved0: __cs__.CharArray | None = ...,
            ClustersPerIndexBuffer: _c_ntfs.int8 | None = ...,
            Reserved1: __cs__.CharArray | None = ...,
            SerialNumber: _c_ntfs.uint64 | None = ...,
            Checksum: _c_ntfs.uint32 | None = ...,
            BootStrap: __cs__.CharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    BOOT_SECTOR: TypeAlias = _BOOT_SECTOR
    class _MFT_SEGMENT_REFERENCE(__cs__.Structure):
        SegmentNumberLowPart: _c_ntfs.uint32
        SegmentNumberHighPart: _c_ntfs.uint16
        SequenceNumber: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            SegmentNumberLowPart: _c_ntfs.uint32 | None = ...,
            SegmentNumberHighPart: _c_ntfs.uint16 | None = ...,
            SequenceNumber: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    MFT_SEGMENT_REFERENCE: TypeAlias = _MFT_SEGMENT_REFERENCE
    FILE_REFERENCE: TypeAlias = _MFT_SEGMENT_REFERENCE
    class _MULTI_SECTOR_HEADER(__cs__.Structure):
        Signature: __cs__.CharArray
        UpdateSequenceArrayOffset: _c_ntfs.uint16
        UpdateSequenceArraySize: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            Signature: __cs__.CharArray | None = ...,
            UpdateSequenceArrayOffset: _c_ntfs.uint16 | None = ...,
            UpdateSequenceArraySize: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    MULTI_SECTOR_HEADER: TypeAlias = _MULTI_SECTOR_HEADER
    class _FILE_RECORD_SEGMENT_HEADER(__cs__.Structure):
        MultiSectorHeader: _c_ntfs._MULTI_SECTOR_HEADER
        Lsn: _c_ntfs.uint64
        SequenceNumber: _c_ntfs.uint16
        ReferenceCount: _c_ntfs.uint16
        FirstAttributeOffset: _c_ntfs.uint16
        Flags: _c_ntfs.uint16
        BytesInUse: _c_ntfs.uint32
        BytesAllocated: _c_ntfs.uint32
        BaseFileRecordSegment: _c_ntfs._MFT_SEGMENT_REFERENCE
        NextAttributeInstance: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            MultiSectorHeader: _c_ntfs._MULTI_SECTOR_HEADER | None = ...,
            Lsn: _c_ntfs.uint64 | None = ...,
            SequenceNumber: _c_ntfs.uint16 | None = ...,
            ReferenceCount: _c_ntfs.uint16 | None = ...,
            FirstAttributeOffset: _c_ntfs.uint16 | None = ...,
            Flags: _c_ntfs.uint16 | None = ...,
            BytesInUse: _c_ntfs.uint32 | None = ...,
            BytesAllocated: _c_ntfs.uint32 | None = ...,
            BaseFileRecordSegment: _c_ntfs._MFT_SEGMENT_REFERENCE | None = ...,
            NextAttributeInstance: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    FILE_RECORD_SEGMENT_HEADER: TypeAlias = _FILE_RECORD_SEGMENT_HEADER
    class ATTRIBUTE_TYPE_CODE(__cs__.Enum):
        UNUSED = ...
        STANDARD_INFORMATION = ...
        ATTRIBUTE_LIST = ...
        FILE_NAME = ...
        OBJECT_ID = ...
        SECURITY_DESCRIPTOR = ...
        VOLUME_NAME = ...
        VOLUME_INFORMATION = ...
        DATA = ...
        INDEX_ROOT = ...
        INDEX_ALLOCATION = ...
        BITMAP = ...
        REPARSE_POINT = ...
        EA_INFORMATION = ...
        EA = ...
        PROPERTY_SET = ...
        LOGGED_UTILITY_STREAM = ...
        END = ...

    class _ATTRIBUTE_RECORD_HEADER(__cs__.Structure):
        TypeCode: _c_ntfs.ATTRIBUTE_TYPE_CODE
        RecordLength: _c_ntfs.uint32
        FormCode: _c_ntfs.uint8
        NameLength: _c_ntfs.uint8
        NameOffset: _c_ntfs.uint16
        Flags: _c_ntfs.uint16
        Instance: _c_ntfs.uint16
        class __anonymous_2__(__cs__.Union):
            class __anonymous_0__(__cs__.Structure):
                ValueLength: _c_ntfs.uint32
                ValueOffset: _c_ntfs.uint16
                Flags: _c_ntfs.uint8
                Reserved: _c_ntfs.uint8
                @overload
                def __init__(
                    self,
                    ValueLength: _c_ntfs.uint32 | None = ...,
                    ValueOffset: _c_ntfs.uint16 | None = ...,
                    Flags: _c_ntfs.uint8 | None = ...,
                    Reserved: _c_ntfs.uint8 | None = ...,
                ): ...
                @overload
                def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

            Resident: __anonymous_0__
            class __anonymous_1__(__cs__.Structure):
                LowestVcn: _c_ntfs.uint64
                HighestVcn: _c_ntfs.uint64
                MappingPairsOffset: _c_ntfs.uint16
                CompressionUnit: _c_ntfs.uint8
                Reserved: __cs__.Array[_c_ntfs.uint8]
                AllocatedLength: _c_ntfs.int64
                FileSize: _c_ntfs.int64
                ValidDataLength: _c_ntfs.int64
                TotalAllocated: _c_ntfs.int64
                @overload
                def __init__(
                    self,
                    LowestVcn: _c_ntfs.uint64 | None = ...,
                    HighestVcn: _c_ntfs.uint64 | None = ...,
                    MappingPairsOffset: _c_ntfs.uint16 | None = ...,
                    CompressionUnit: _c_ntfs.uint8 | None = ...,
                    Reserved: __cs__.Array[_c_ntfs.uint8] | None = ...,
                    AllocatedLength: _c_ntfs.int64 | None = ...,
                    FileSize: _c_ntfs.int64 | None = ...,
                    ValidDataLength: _c_ntfs.int64 | None = ...,
                    TotalAllocated: _c_ntfs.int64 | None = ...,
                ): ...
                @overload
                def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

            Nonresident: __anonymous_1__
            @overload
            def __init__(self, Resident: __anonymous_0__ | None = ..., Nonresident: __anonymous_1__ | None = ...): ...
            @overload
            def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

        Form: __anonymous_2__
        @overload
        def __init__(
            self,
            TypeCode: _c_ntfs.ATTRIBUTE_TYPE_CODE | None = ...,
            RecordLength: _c_ntfs.uint32 | None = ...,
            FormCode: _c_ntfs.uint8 | None = ...,
            NameLength: _c_ntfs.uint8 | None = ...,
            NameOffset: _c_ntfs.uint16 | None = ...,
            Flags: _c_ntfs.uint16 | None = ...,
            Instance: _c_ntfs.uint16 | None = ...,
            Form: __anonymous_2__ | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    ATTRIBUTE_RECORD_HEADER: TypeAlias = _ATTRIBUTE_RECORD_HEADER
    class _STANDARD_INFORMATION(__cs__.Structure):
        CreationTime: _c_ntfs.int64
        LastModificationTime: _c_ntfs.int64
        LastChangeTime: _c_ntfs.int64
        LastAccessTime: _c_ntfs.int64
        FileAttributes: _c_ntfs.uint32
        MaximumVersions: _c_ntfs.uint32
        VersionNumber: _c_ntfs.uint32
        ClassId: _c_ntfs.uint32
        OwnerId: _c_ntfs.uint32
        SecurityId: _c_ntfs.uint32
        QuotaCharged: _c_ntfs.uint64
        Usn: _c_ntfs.uint64
        @overload
        def __init__(
            self,
            CreationTime: _c_ntfs.int64 | None = ...,
            LastModificationTime: _c_ntfs.int64 | None = ...,
            LastChangeTime: _c_ntfs.int64 | None = ...,
            LastAccessTime: _c_ntfs.int64 | None = ...,
            FileAttributes: _c_ntfs.uint32 | None = ...,
            MaximumVersions: _c_ntfs.uint32 | None = ...,
            VersionNumber: _c_ntfs.uint32 | None = ...,
            ClassId: _c_ntfs.uint32 | None = ...,
            OwnerId: _c_ntfs.uint32 | None = ...,
            SecurityId: _c_ntfs.uint32 | None = ...,
            QuotaCharged: _c_ntfs.uint64 | None = ...,
            Usn: _c_ntfs.uint64 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    STANDARD_INFORMATION_EX: TypeAlias = _STANDARD_INFORMATION
    class _ATTRIBUTE_LIST_ENTRY(__cs__.Structure):
        AttributeTypeCode: _c_ntfs.ATTRIBUTE_TYPE_CODE
        RecordLength: _c_ntfs.uint16
        AttributeNameLength: _c_ntfs.uint8
        AttributeNameOffset: _c_ntfs.uint8
        LowestVcn: _c_ntfs.uint64
        SegmentReference: _c_ntfs._MFT_SEGMENT_REFERENCE
        Reserved: _c_ntfs.uint16
        AttributeName: __cs__.WcharArray
        @overload
        def __init__(
            self,
            AttributeTypeCode: _c_ntfs.ATTRIBUTE_TYPE_CODE | None = ...,
            RecordLength: _c_ntfs.uint16 | None = ...,
            AttributeNameLength: _c_ntfs.uint8 | None = ...,
            AttributeNameOffset: _c_ntfs.uint8 | None = ...,
            LowestVcn: _c_ntfs.uint64 | None = ...,
            SegmentReference: _c_ntfs._MFT_SEGMENT_REFERENCE | None = ...,
            Reserved: _c_ntfs.uint16 | None = ...,
            AttributeName: __cs__.WcharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    ATTRIBUTE_LIST_ENTRY: TypeAlias = _ATTRIBUTE_LIST_ENTRY
    class _FILE_NAME(__cs__.Structure):
        ParentDirectory: _c_ntfs._MFT_SEGMENT_REFERENCE
        CreationTime: _c_ntfs.int64
        LastModificationTime: _c_ntfs.int64
        LastChangeTime: _c_ntfs.int64
        LastAccessTime: _c_ntfs.int64
        AllocatedLength: _c_ntfs.int64
        FileSize: _c_ntfs.int64
        FileAttributes: _c_ntfs.uint32
        EaSize: _c_ntfs.uint16
        _: _c_ntfs.uint16
        ReparsePointTag: _c_ntfs.uint32
        FileNameLength: _c_ntfs.uint8
        Flags: _c_ntfs.uint8
        FileName: __cs__.WcharArray
        @overload
        def __init__(
            self,
            ParentDirectory: _c_ntfs._MFT_SEGMENT_REFERENCE | None = ...,
            CreationTime: _c_ntfs.int64 | None = ...,
            LastModificationTime: _c_ntfs.int64 | None = ...,
            LastChangeTime: _c_ntfs.int64 | None = ...,
            LastAccessTime: _c_ntfs.int64 | None = ...,
            AllocatedLength: _c_ntfs.int64 | None = ...,
            FileSize: _c_ntfs.int64 | None = ...,
            FileAttributes: _c_ntfs.uint32 | None = ...,
            EaSize: _c_ntfs.uint16 | None = ...,
            _: _c_ntfs.uint16 | None = ...,
            ReparsePointTag: _c_ntfs.uint32 | None = ...,
            FileNameLength: _c_ntfs.uint8 | None = ...,
            Flags: _c_ntfs.uint8 | None = ...,
            FileName: __cs__.WcharArray | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    FILE_NAME: TypeAlias = _FILE_NAME
    class IO_REPARSE_TAG(__cs__.Enum):
        RESERVED_ZERO = ...
        RESERVED_ONE = ...
        RESERVED_TWO = ...
        MOUNT_POINT = ...
        HSM = ...
        DRIVE_EXTENDER = ...
        HSM2 = ...
        SIS = ...
        WIM = ...
        CSV = ...
        DFS = ...
        FILTER_MANAGER = ...
        SYMLINK = ...
        IIS_CACHE = ...
        DFSR = ...
        DEDUP = ...
        APPXSTRM = ...
        NFS = ...
        FILE_PLACEHOLDER = ...
        DFM = ...
        WOF = ...
        WCI = ...
        WCI_1 = ...
        GLOBAL_REPARSE = ...
        CLOUD = ...
        CLOUD_1 = ...
        CLOUD_2 = ...
        CLOUD_3 = ...
        CLOUD_4 = ...
        CLOUD_5 = ...
        CLOUD_6 = ...
        CLOUD_7 = ...
        CLOUD_8 = ...
        CLOUD_9 = ...
        CLOUD_A = ...
        CLOUD_B = ...
        CLOUD_C = ...
        CLOUD_D = ...
        CLOUD_E = ...
        CLOUD_F = ...
        APPEXECLINK = ...
        PROJFS = ...
        LX_SYMLINK = ...
        STORAGE_SYNC = ...
        WCI_TOMBSTONE = ...
        UNHANDLED = ...
        ONEDRIVE = ...
        PROJFS_TOMBSTONE = ...
        AF_UNIX = ...
        LX_FIFO = ...
        LX_CHR = ...
        LX_BLK = ...
        WCI_LINK = ...
        WCI_LINK_1 = ...

    class _REPARSE_DATA_BUFFER(__cs__.Structure):
        ReparseTag: _c_ntfs.IO_REPARSE_TAG
        ReparseDataLength: _c_ntfs.uint16
        Reserved: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            ReparseTag: _c_ntfs.IO_REPARSE_TAG | None = ...,
            ReparseDataLength: _c_ntfs.uint16 | None = ...,
            Reserved: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    REPARSE_DATA_BUFFER: TypeAlias = _REPARSE_DATA_BUFFER
    class SYMLINK_FLAG(__cs__.Enum):
        ABSOLUTE = ...
        RELATIVE = ...

    class _SYMBOLIC_LINK_REPARSE_BUFFER(__cs__.Structure):
        SubstituteNameOffset: _c_ntfs.uint16
        SubstituteNameLength: _c_ntfs.uint16
        PrintNameOffset: _c_ntfs.uint16
        PrintNameLength: _c_ntfs.uint16
        Flags: _c_ntfs.SYMLINK_FLAG
        @overload
        def __init__(
            self,
            SubstituteNameOffset: _c_ntfs.uint16 | None = ...,
            SubstituteNameLength: _c_ntfs.uint16 | None = ...,
            PrintNameOffset: _c_ntfs.uint16 | None = ...,
            PrintNameLength: _c_ntfs.uint16 | None = ...,
            Flags: _c_ntfs.SYMLINK_FLAG | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    SYMBOLIC_LINK_REPARSE_BUFFER: TypeAlias = _SYMBOLIC_LINK_REPARSE_BUFFER
    class _MOUNT_POINT_REPARSE_BUFFER(__cs__.Structure):
        SubstituteNameOffset: _c_ntfs.uint16
        SubstituteNameLength: _c_ntfs.uint16
        PrintNameOffset: _c_ntfs.uint16
        PrintNameLength: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            SubstituteNameOffset: _c_ntfs.uint16 | None = ...,
            SubstituteNameLength: _c_ntfs.uint16 | None = ...,
            PrintNameOffset: _c_ntfs.uint16 | None = ...,
            PrintNameLength: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class COLLATION(__cs__.Enum):
        BINARY = ...
        FILE_NAME = ...
        UNICODE_STRING = ...
        NUMBER_RULES = ...
        NTOFS_ULONG = ...
        NTOFS_SID = ...
        NTOFS_SECURITY_HASH = ...
        NTOFS_ULONGS = ...

    COLLATION_RULE: TypeAlias = COLLATION
    class _INDEX_HEADER(__cs__.Structure):
        FirstEntryOffset: _c_ntfs.uint32
        TotalSizeOfEntries: _c_ntfs.uint32
        AllocatedSize: _c_ntfs.uint32
        Flags: _c_ntfs.uint8
        Reserved: __cs__.Array[_c_ntfs.uint8]
        @overload
        def __init__(
            self,
            FirstEntryOffset: _c_ntfs.uint32 | None = ...,
            TotalSizeOfEntries: _c_ntfs.uint32 | None = ...,
            AllocatedSize: _c_ntfs.uint32 | None = ...,
            Flags: _c_ntfs.uint8 | None = ...,
            Reserved: __cs__.Array[_c_ntfs.uint8] | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    INDEX_HEADER: TypeAlias = _INDEX_HEADER
    class _INDEX_ROOT(__cs__.Structure):
        AttributeType: _c_ntfs.ATTRIBUTE_TYPE_CODE
        CollationRule: _c_ntfs.COLLATION
        BytesPerIndexBuffer: _c_ntfs.uint32
        ClustersPerIndexBuffer: _c_ntfs.uint8
        Reserved: __cs__.Array[_c_ntfs.uint8]
        IndexHeader: _c_ntfs._INDEX_HEADER
        @overload
        def __init__(
            self,
            AttributeType: _c_ntfs.ATTRIBUTE_TYPE_CODE | None = ...,
            CollationRule: _c_ntfs.COLLATION | None = ...,
            BytesPerIndexBuffer: _c_ntfs.uint32 | None = ...,
            ClustersPerIndexBuffer: _c_ntfs.uint8 | None = ...,
            Reserved: __cs__.Array[_c_ntfs.uint8] | None = ...,
            IndexHeader: _c_ntfs._INDEX_HEADER | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    INDEX_ROOT: TypeAlias = _INDEX_ROOT
    class _INDEX_ALLOCATION_BUFFER(__cs__.Structure):
        MultiSectorHeader: _c_ntfs._MULTI_SECTOR_HEADER
        Lsn: _c_ntfs.uint64
        Vcn: _c_ntfs.uint64
        IndexHeader: _c_ntfs._INDEX_HEADER
        @overload
        def __init__(
            self,
            MultiSectorHeader: _c_ntfs._MULTI_SECTOR_HEADER | None = ...,
            Lsn: _c_ntfs.uint64 | None = ...,
            Vcn: _c_ntfs.uint64 | None = ...,
            IndexHeader: _c_ntfs._INDEX_HEADER | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    INDEX_ALLOCATION_BUFFER: TypeAlias = _INDEX_ALLOCATION_BUFFER
    class _INDEX_ENTRY(__cs__.Structure):
        FileReference: _c_ntfs._MFT_SEGMENT_REFERENCE
        DataOffset: _c_ntfs.uint16
        DataLength: _c_ntfs.uint16
        _: _c_ntfs.uint32
        Length: _c_ntfs.uint16
        KeyLength: _c_ntfs.uint16
        Flags: _c_ntfs.uint16
        Reserved: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            FileReference: _c_ntfs._MFT_SEGMENT_REFERENCE | None = ...,
            DataOffset: _c_ntfs.uint16 | None = ...,
            DataLength: _c_ntfs.uint16 | None = ...,
            _: _c_ntfs.uint32 | None = ...,
            Length: _c_ntfs.uint16 | None = ...,
            KeyLength: _c_ntfs.uint16 | None = ...,
            Flags: _c_ntfs.uint16 | None = ...,
            Reserved: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    INDEX_ENTRY: TypeAlias = _INDEX_ENTRY
    class SECURITY_DESCRIPTOR_CONTROL(__cs__.Flag):
        SE_OWNER_DEFAULTED = ...
        SE_GROUP_DEFAULTED = ...
        SE_DACL_PRESENT = ...
        SE_DACL_DEFAULTED = ...
        SE_SACL_PRESENT = ...
        SE_SACL_DEFAULTED = ...
        SE_DACL_AUTO_INHERIT_REQ = ...
        SE_SACL_AUTO_INHERIT_REQ = ...
        SE_DACL_AUTO_INHERITED = ...
        SE_SACL_AUTO_INHERITED = ...
        SE_DACL_PROTECTED = ...
        SE_SACL_PROTECTED = ...
        SE_RM_CONTROL_VALID = ...
        SE_SELF_RELATIVE = ...

    class ACCESS_MASK(__cs__.Flag):
        FILE_READ_DATA = ...
        FILE_LIST_DIRECTORY = ...
        FILE_WRITE_DATA = ...
        FILE_ADD_FILE = ...
        FILE_APPEND_DATA = ...
        FILE_ADD_SUBDIRECTORY = ...
        FILE_READ_EA = ...
        FILE_WRITE_EA = ...
        FILE_EXECUTE = ...
        FILE_TRAVERSE = ...
        FILE_DELETE_CHILD = ...
        FILE_READ_ATTRIBUTES = ...
        FILE_WRITE_ATTRIBUTES = ...
        DELETE = ...
        READ_CONTROL = ...
        WRITE_DAC = ...
        WRITE_OWNER = ...
        SYNCHRONIZE = ...
        STANDARD_RIGHTS_READ = ...
        STANDARD_RIGHTS_WRITE = ...
        STANDARD_RIGHTS_EXECUTE = ...
        STANDARD_RIGHTS_REQUIRED = ...
        STANDARD_RIGHTS_ALL = ...
        ACCESS_SYSTEM_SECURITY = ...
        MAXIMUM_ALLOWED = ...
        GENERIC_ALL = ...
        GENERIC_EXECUTE = ...
        GENERIC_WRITE = ...
        GENERIC_READ = ...

    class ACE_TYPE(__cs__.Enum):
        ACCESS_ALLOWED = ...
        ACCESS_DENIED = ...
        SYSTEM_AUDIT = ...
        SYSTEM_ALARM = ...
        ACCESS_ALLOWED_COMPOUND = ...
        ACCESS_ALLOWED_OBJECT = ...
        ACCESS_DENIED_OBJECT = ...
        SYSTEM_AUDIT_OBJECT = ...
        SYSTEM_ALARM_OBJECT = ...
        ACCESS_ALLOWED_CALLBACK = ...
        ACCESS_DENIED_CALLBACK = ...
        ACCESS_ALLOWED_CALLBACK_OBJECT = ...
        ACCESS_DENIED_CALLBACK_OBJECT = ...
        SYSTEM_AUDIT_CALLBACK = ...
        SYSTEM_ALARM_CALLBACK = ...
        SYSTEM_AUDIT_CALLBACK_OBJECT = ...
        SYSTEM_ALARM_CALLBACK_OBJECT = ...
        SYSTEM_MANDATORY_LABEL = ...
        SYSTEM_RESOURCE_ATTRIBUTE = ...
        SYSTEM_SCOPED_POLICY_ID = ...
        SYSTEM_PROCESS_TRUST_LABEL = ...
        SYSTEM_ACCESS_FILTER = ...

    class ACE_FLAGS(__cs__.Flag):
        OBJECT_INHERIT_ACE = ...
        CONTAINER_INHERIT_ACE = ...
        NO_PROPAGATE_INHERIT_ACE = ...
        INHERIT_ONLY_ACE = ...
        INHERITED_ACE = ...
        SUCCESSFUL_ACCESS_ACE_FLAG = ...
        FAILED_ACCESS_ACE_FLAG = ...

    class ACE_OBJECT_FLAGS(__cs__.Flag):
        ACE_OBJECT_TYPE_PRESENT = ...
        ACE_INHERITED_OBJECT_TYPE_PRESENT = ...

    class COMPOUND_ACE_TYPE(__cs__.Enum):
        COMPOUND_ACE_IMPERSONATION = ...

    class _ACL(__cs__.Structure):
        AclRevision: _c_ntfs.uint8
        Sbz1: _c_ntfs.uint8
        AclSize: _c_ntfs.uint16
        AceCount: _c_ntfs.uint16
        Sbz2: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            AclRevision: _c_ntfs.uint8 | None = ...,
            Sbz1: _c_ntfs.uint8 | None = ...,
            AclSize: _c_ntfs.uint16 | None = ...,
            AceCount: _c_ntfs.uint16 | None = ...,
            Sbz2: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    ACL: TypeAlias = _ACL
    class _ACE_HEADER(__cs__.Structure):
        AceType: _c_ntfs.ACE_TYPE
        AceFlags: _c_ntfs.ACE_FLAGS
        AceSize: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            AceType: _c_ntfs.ACE_TYPE | None = ...,
            AceFlags: _c_ntfs.ACE_FLAGS | None = ...,
            AceSize: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    ACE_HEADER: TypeAlias = _ACE_HEADER
    class _SECURITY_DESCRIPTOR_HEADER(__cs__.Structure):
        HashId: _c_ntfs.uint32
        SecurityId: _c_ntfs.uint32
        Offset: _c_ntfs.uint64
        Length: _c_ntfs.uint32
        @overload
        def __init__(
            self,
            HashId: _c_ntfs.uint32 | None = ...,
            SecurityId: _c_ntfs.uint32 | None = ...,
            Offset: _c_ntfs.uint64 | None = ...,
            Length: _c_ntfs.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    SECURITY_DESCRIPTOR_HEADER: TypeAlias = _SECURITY_DESCRIPTOR_HEADER
    class _SECURITY_DESCRIPTOR_RELATIVE(__cs__.Structure):
        Revision: _c_ntfs.uint8
        Sbz1: _c_ntfs.uint8
        Control: _c_ntfs.SECURITY_DESCRIPTOR_CONTROL
        Owner: _c_ntfs.uint32
        Group: _c_ntfs.uint32
        Sacl: _c_ntfs.uint32
        Dacl: _c_ntfs.uint32
        @overload
        def __init__(
            self,
            Revision: _c_ntfs.uint8 | None = ...,
            Sbz1: _c_ntfs.uint8 | None = ...,
            Control: _c_ntfs.SECURITY_DESCRIPTOR_CONTROL | None = ...,
            Owner: _c_ntfs.uint32 | None = ...,
            Group: _c_ntfs.uint32 | None = ...,
            Sacl: _c_ntfs.uint32 | None = ...,
            Dacl: _c_ntfs.uint32 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    SECURITY_DESCRIPTOR_RELATIVE: TypeAlias = _SECURITY_DESCRIPTOR_RELATIVE
    class USN_REASON(__cs__.Flag):
        DATA_OVERWRITE = ...
        DATA_EXTEND = ...
        DATA_TRUNCATION = ...
        NAMED_DATA_OVERWRITE = ...
        NAMED_DATA_EXTEND = ...
        NAMED_DATA_TRUNCATION = ...
        FILE_CREATE = ...
        FILE_DELETE = ...
        EA_CHANGE = ...
        SECURITY_CHANGE = ...
        RENAME_OLD_NAME = ...
        RENAME_NEW_NAME = ...
        INDEXABLE_CHANGE = ...
        BASIC_INFO_CHANGE = ...
        HARD_LINK_CHANGE = ...
        COMPRESSION_CHANGE = ...
        ENCRYPTION_CHANGE = ...
        OBJECT_ID_CHANGE = ...
        REPARSE_POINT_CHANGE = ...
        STREAM_CHANGE = ...
        TRANSACTED_CHANGE = ...
        INTEGRITY_CHANGE = ...
        CLOSE = ...

    class USN_SOURCE(__cs__.Flag):
        NORMAL = ...
        DATA_MANAGEMENT = ...
        AUXILIARY_DATA = ...
        REPLICATION_MANAGEMENT = ...
        CLIENT_REPLICATION_MANAGEMENT = ...

    class _FILE_ID_128(__cs__.Structure):
        Identifier: __cs__.Array[_c_ntfs.uint8]
        @overload
        def __init__(self, Identifier: __cs__.Array[_c_ntfs.uint8] | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    FILE_ID_128: TypeAlias = _FILE_ID_128
    class USN_RECORD_COMMON_HEADER(__cs__.Structure):
        RecordLength: _c_ntfs.uint32
        MajorVersion: _c_ntfs.uint16
        MinorVersion: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            RecordLength: _c_ntfs.uint32 | None = ...,
            MajorVersion: _c_ntfs.uint16 | None = ...,
            MinorVersion: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class USN_RECORD_V2(__cs__.Structure):
        RecordLength: _c_ntfs.uint32
        MajorVersion: _c_ntfs.uint16
        MinorVersion: _c_ntfs.uint16
        FileReferenceNumber: _c_ntfs._MFT_SEGMENT_REFERENCE
        ParentFileReferenceNumber: _c_ntfs._MFT_SEGMENT_REFERENCE
        Usn: _c_ntfs.uint64
        TimeStamp: _c_ntfs.uint64
        Reason: _c_ntfs.USN_REASON
        SourceInfo: _c_ntfs.USN_SOURCE
        SecurityId: _c_ntfs.uint32
        FileAttributes: _c_ntfs.FILE_ATTRIBUTE
        FileNameLength: _c_ntfs.uint16
        FileNameOffset: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            RecordLength: _c_ntfs.uint32 | None = ...,
            MajorVersion: _c_ntfs.uint16 | None = ...,
            MinorVersion: _c_ntfs.uint16 | None = ...,
            FileReferenceNumber: _c_ntfs._MFT_SEGMENT_REFERENCE | None = ...,
            ParentFileReferenceNumber: _c_ntfs._MFT_SEGMENT_REFERENCE | None = ...,
            Usn: _c_ntfs.uint64 | None = ...,
            TimeStamp: _c_ntfs.uint64 | None = ...,
            Reason: _c_ntfs.USN_REASON | None = ...,
            SourceInfo: _c_ntfs.USN_SOURCE | None = ...,
            SecurityId: _c_ntfs.uint32 | None = ...,
            FileAttributes: _c_ntfs.FILE_ATTRIBUTE | None = ...,
            FileNameLength: _c_ntfs.uint16 | None = ...,
            FileNameOffset: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class USN_RECORD_V3(__cs__.Structure):
        RecordLength: _c_ntfs.uint32
        MajorVersion: _c_ntfs.uint16
        MinorVersion: _c_ntfs.uint16
        FileReferenceNumber: _c_ntfs._FILE_ID_128
        ParentFileReferenceNumber: _c_ntfs._FILE_ID_128
        Usn: _c_ntfs.uint64
        TimeStamp: _c_ntfs.uint64
        Reason: _c_ntfs.USN_REASON
        SourceInfo: _c_ntfs.USN_SOURCE
        SecurityId: _c_ntfs.uint32
        FileAttributes: _c_ntfs.FILE_ATTRIBUTE
        FileNameLength: _c_ntfs.uint16
        FileNameOffset: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            RecordLength: _c_ntfs.uint32 | None = ...,
            MajorVersion: _c_ntfs.uint16 | None = ...,
            MinorVersion: _c_ntfs.uint16 | None = ...,
            FileReferenceNumber: _c_ntfs._FILE_ID_128 | None = ...,
            ParentFileReferenceNumber: _c_ntfs._FILE_ID_128 | None = ...,
            Usn: _c_ntfs.uint64 | None = ...,
            TimeStamp: _c_ntfs.uint64 | None = ...,
            Reason: _c_ntfs.USN_REASON | None = ...,
            SourceInfo: _c_ntfs.USN_SOURCE | None = ...,
            SecurityId: _c_ntfs.uint32 | None = ...,
            FileAttributes: _c_ntfs.FILE_ATTRIBUTE | None = ...,
            FileNameLength: _c_ntfs.uint16 | None = ...,
            FileNameOffset: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class USN_RECORD_EXTENT(__cs__.Structure):
        Offset: _c_ntfs.int64
        Length: _c_ntfs.int64
        @overload
        def __init__(self, Offset: _c_ntfs.int64 | None = ..., Length: _c_ntfs.int64 | None = ...): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

    class USN_RECORD_V4(__cs__.Structure):
        RecordLength: _c_ntfs.uint32
        MajorVersion: _c_ntfs.uint16
        MinorVersion: _c_ntfs.uint16
        FileReferenceNumber: _c_ntfs._FILE_ID_128
        ParentFileReferenceNumber: _c_ntfs._FILE_ID_128
        Usn: _c_ntfs.uint64
        Reason: _c_ntfs.USN_REASON
        SourceInfo: _c_ntfs.USN_SOURCE
        RemainingExtents: _c_ntfs.uint32
        NumberOfExtents: _c_ntfs.uint16
        ExtentSize: _c_ntfs.uint16
        @overload
        def __init__(
            self,
            RecordLength: _c_ntfs.uint32 | None = ...,
            MajorVersion: _c_ntfs.uint16 | None = ...,
            MinorVersion: _c_ntfs.uint16 | None = ...,
            FileReferenceNumber: _c_ntfs._FILE_ID_128 | None = ...,
            ParentFileReferenceNumber: _c_ntfs._FILE_ID_128 | None = ...,
            Usn: _c_ntfs.uint64 | None = ...,
            Reason: _c_ntfs.USN_REASON | None = ...,
            SourceInfo: _c_ntfs.USN_SOURCE | None = ...,
            RemainingExtents: _c_ntfs.uint32 | None = ...,
            NumberOfExtents: _c_ntfs.uint16 | None = ...,
            ExtentSize: _c_ntfs.uint16 | None = ...,
        ): ...
        @overload
        def __init__(self, fh: bytes | memoryview | bytearray | BinaryIO, /): ...

# Technically `c_ntfs` is an instance of `_c_ntfs`, but then we can't use it in type hints
c_ntfs: TypeAlias = _c_ntfs

ATTRIBUTE_TYPE_CODE: TypeAlias = c_ntfs.ATTRIBUTE_TYPE_CODE
IO_REPARSE_TAG: TypeAlias = c_ntfs.IO_REPARSE_TAG
ACCESS_MASK: TypeAlias = c_ntfs.ACCESS_MASK
ACE_TYPE: TypeAlias = c_ntfs.ACE_TYPE
ACE_OBJECT_FLAGS: TypeAlias = c_ntfs.ACE_OBJECT_FLAGS
COLLATION: TypeAlias = c_ntfs.COLLATION

NTFS_SIGNATURE: bytes = ...

SECTOR_SIZE: int = ...
SECTOR_SHIFT: int = ...

USN_PAGE_SIZE: int = ...

DEFAULT_SECTOR_SIZE: int = ...
DEFAULT_CLUSTER_SIZE: int = ...
DEFAULT_RECORD_SIZE: int = ...
DEFAULT_INDEX_SIZE: int = ...

FILE_NUMBER_MFT: int = ...
FILE_NUMBER_MFTMIRR: int = ...
FILE_NUMBER_LOGFILE: int = ...
FILE_NUMBER_VOLUME: int = ...
FILE_NUMBER_ATTRDEF: int = ...
FILE_NUMBER_ROOT: int = ...
FILE_NUMBER_BITMAP: int = ...
FILE_NUMBER_BOOT: int = ...
FILE_NUMBER_BADCLUS: int = ...
FILE_NUMBER_SECURE: int = ...
FILE_NUMBER_UPCASE: int = ...
FILE_NUMBER_EXTEND: int = ...

FILE_RECORD_SEGMENT_IN_USE: int = ...
FILE_FILE_NAME_INDEX_PRESENT: int = ...

ATTRIBUTE_FLAG_COMPRESSION_MASK: int = ...
ATTRIBUTE_FLAG_ENCRYPTED: int = ...
ATTRIBUTE_FLAG_SPARSE: int = ...

FILE_NAME_NTFS: int = ...
FILE_NAME_DOS: int = ...

COMPRESSION_FORMAT_NONE: int = ...
COMPRESSION_FORMAT_DEFAULT: int = ...
COMPRESSION_FORMAT_LZNT1: int = ...

INDEX_NODE: int = ...
INDEX_ENTRY_NODE: int = ...
INDEX_ENTRY_END: int = ...



from __future__ import annotations

import io
import struct
from typing import BinaryIO
from unittest.mock import Mock

from dissect.ntfs.c_ntfs import ATTRIBUTE_TYPE_CODE, c_ntfs
from dissect.ntfs.index import IndexEntry, Match, _cmp_filename, _cmp_ulong
from dissect.ntfs.ntfs import NTFS


def mock_filename_entry(filename: str) -> IndexEntry:
    attribute = c_ntfs._FILE_NAME(
        FileNameLength=len(filename),
        FileName=filename,
    ).dumps()
    header = c_ntfs._INDEX_ENTRY(
        Length=len(c_ntfs._INDEX_ENTRY) + len(attribute),
        KeyLength=len(attribute),
    )
    data = header.dumps() + attribute

    mock_root = Mock(attribute_type=ATTRIBUTE_TYPE_CODE.FILE_NAME)
    mock_index = Mock(root=mock_root)
    return IndexEntry(mock_index, io.BytesIO(data), 0)


def mock_ulong_entry(value: int) -> IndexEntry:
    header = c_ntfs._INDEX_ENTRY(
        Length=len(c_ntfs._INDEX_ENTRY) + 4,
        KeyLength=4,
    )
    data = header.dumps() + struct.pack("<I", value)

    mock_root = Mock(attribute_type=ATTRIBUTE_TYPE_CODE.UNUSED)
    mock_index = Mock(root=mock_root)
    return IndexEntry(mock_index, io.BytesIO(data), 0)


def test_cmp_filename() -> None:
    entry = mock_filename_entry("bbbb")

    assert _cmp_filename(entry, "CCCC") == Match.Greater
    assert _cmp_filename(entry, "BBBB") == Match.Equal
    assert _cmp_filename(entry, "BBBA") == Match.Less
    assert _cmp_filename(entry, "BBBBA") == Match.Greater
    assert _cmp_filename(entry, "BBBBB") == Match.Greater

    entry = mock_filename_entry("C_20127.NLS")
    assert _cmp_filename(entry, "CONFIG") == Match.Less


def test_cmp_ulong() -> None:
    entry = mock_ulong_entry(100)

    assert _cmp_ulong(entry, 99) == Match.Less
    assert _cmp_ulong(entry, 100) == Match.Equal
    assert _cmp_ulong(entry, 101) == Match.Greater


def test_index_lookup(ntfs_bin: BinaryIO) -> None:
    fs = NTFS(ntfs_bin)

    root = fs.mft.get("Large Directory")
    assert len(root.listdir()) == 2048

    for i in range(2048):
        entry = root.get(f"Directory {i}")
        assert entry.filename == f"Directory {i}"



from __future__ import annotations

from io import BytesIO
from typing import BinaryIO

import pytest

from dissect.ntfs.exceptions import FileNotFoundError, NotADirectoryError
from dissect.ntfs.ntfs import NTFS


def test_ntfs(ntfs_bin: BinaryIO) -> None:
    fs = NTFS(ntfs_bin)

    assert fs.sector_size == 512
    assert fs.cluster_size == 4096
    assert fs._record_size == 1024
    assert fs._index_size == 4096
    assert fs.volume_name == "New Volume"

    small_file = fs.mft.get("File.txt")
    assert small_file.segment == 41
    assert small_file.resident
    assert small_file.is_file()
    assert not small_file.is_dir()
    assert small_file.filename == "File.txt"
    assert small_file.size() == 13
    assert small_file.open().read() == b"Contents here"

    with pytest.raises(FileNotFoundError):
        small_file.index("$I30")

    large_file = fs.mft.get("Large.txt")
    assert large_file.segment == 44
    assert not large_file.resident
    assert large_file.is_file()
    assert not large_file.is_dir()
    assert large_file.filename == "Large.txt"
    assert large_file.size() == 2097152
    assert large_file.open().read() == 2097152 * b"A"

    with pytest.raises(NotADirectoryError):
        large_file.get("Something")

    directory = fs.mft.get("Directory")
    assert not directory.is_file()
    assert directory.is_dir()
    assert directory.index("$I30")
    assert list(directory.listdir().keys()) == ["File 1.txt", "File 2.txt"]

    assert directory.get("File 1.txt") == fs.mft.get("Directory/File 1.txt")
    assert fs.mft.get("Directory/File 1.txt").full_path() == "Directory\\File 1.txt"


def test_ntfs_large_sector(boot_2m_bin: BinaryIO) -> None:
    fs = NTFS(boot=boot_2m_bin)
    assert fs.cluster_size == 0x200000


def test_ntfs_64k_sector() -> None:
    boot_sector = """
    eb52904e5446532020202000028000000000000000f800003f00ff0000080400
    0000000080008000ffef3b060000000000c00000000000000100000000000000
    f6000000f4000000dcf8511e0a521e0000000000fa33c08ed0bc007cfb68c007
    1f1e686600cb88160e0066813e03004e5446537515b441bbaa55cd13720c81fb
    55aa7506f7c101007503e9dd001e83ec18681a00b4488a160e008bf4161fcd13
    9f83c4189e581f72e13b060b0075dba30f00c12e0f00041e5a33dbb900202bc8
    66ff06110003160f008ec2ff061600e84b002bc877efb800bbcd1a6623c0752d
    6681fb54435041752481f90201721e166807bb16685211166809006653665366
    5516161668b80166610e07cd1a33c0bf0a13b9f60cfcf3aae9fe01909066601e
    0666a111006603061c001e66680000000066500653680100681000b4428a160e
    00161f8bf4cd1366595b5a665966591f0f82160066ff06110003160f008ec2ff
    0e160075bc071f6661c3a1f601e80900a1fa01e80300f4ebfd8bf0ac3c007409
    b40ebb0700cd10ebf2c30d0a41206469736b2072656164206572726f72206f63
    637572726564000d0a424f4f544d475220697320636f6d70726573736564000d
    0a5072657373204374726c2b416c742b44656c20746f20726573746172740d0a
    000000000000000000000000000000000000000000008a01a701bf01000055aa
    """

    fs = NTFS(boot=BytesIO(bytes.fromhex(boot_sector)))
    assert fs.cluster_size == 0x10000


def test_fragmented_mft(ntfs_fragmented_mft_fh: BinaryIO) -> None:
    fs = NTFS(ntfs_fragmented_mft_fh)
    assert len(fs.mft.fh.runlist) == 238



from __future__ import annotations

import struct
from collections import UserDict
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct import Enum
from dissect.util.stream import RunlistStream

from dissect.ntfs.c_ntfs import (
    ATTRIBUTE_FLAG_COMPRESSION_MASK,
    ATTRIBUTE_TYPE_CODE,
    FILE_NUMBER_ROOT,
    SECTOR_SHIFT,
    SECTOR_SIZE,
    c_ntfs,
)
from dissect.ntfs.exceptions import FilenameNotAvailableError, VolumeNotAvailableError
from dissect.ntfs.stream import CompressedRunlistStream

if TYPE_CHECKING:
    from dissect.ntfs.attr import Attribute
    from dissect.ntfs.mft import Mft
    from dissect.ntfs.ntfs import NTFS


class AttributeMap(UserDict):
    """Utility dictionary-like object for interacting with a collection of attributes.

    Allows convenient accessing of attributes added to this collection. For example:
    - Get attributes by name, e.g. ``attributes.DATA`` to get all ``$DATA`` attributes.
    - Get attributes by type code enum or integer, e.g. ``attributes[0x80]`` or ``attributes[ATTRIBUTE_TYPE_CODE.DATA]``.
    - Check attribute membership by enum or integer, e.g. ``0x80 in attributes`` or ``ATTRIBUTE_TYPE_CODE.DATA in attributes``.
    - Find all attributes with a given name and type, e.g. ``attributes.find("$I30", ATTRIBUTE_TYPE_CODE.INDEX_ROOT)``.

    Note that any data retrieval from an ``AttributeMap`` will always succeed and return an
    :class:`~dissect.ntfs.util.AttributeCollection`, either empty or containing one or more attributes.
    """  # noqa: E501

    def __getattr__(self, attr: str) -> AttributeCollection:
        if attr in ATTRIBUTE_TYPE_CODE.__members__:
            return self[ATTRIBUTE_TYPE_CODE[attr]]

        return super().__getattribute__(attr)

    def __getitem__(self, item: ATTRIBUTE_TYPE_CODE | int) -> AttributeCollection:
        if isinstance(item, Enum):
            item = item.value
        return self.data.get(item, AttributeCollection())

    def __contains__(self, key: ATTRIBUTE_TYPE_CODE | int) -> bool:
        if isinstance(key, Enum):
            key = key.value
        return super().__contains__(key)

    def add(self, attr: Attribute) -> None:
        """Add an attribute to the collection.

        Note that this is the only intended way to modify the :class:`AttributeMap`!

        Args:
            attr: The attribute to add.
        """
        key = attr.header.type.value
        if key not in self:
            self[key] = AttributeCollection()

        self[key].append(attr)

    def find(self, name: str, attr_type: ATTRIBUTE_TYPE_CODE) -> AttributeCollection:
        """Find attributes by name and attribute type.

        Args:
            name: The name of the attribute to find, usually ``""``.
            attr_type: The attribute type to find.
        """
        name = name.lower()
        return AttributeCollection(
            [attr for attr in self.get(attr_type, AttributeCollection()) if attr.header.name.lower() == name]
        )


class AttributeCollection(list):
    """Utility list-like object for interacting with a list of attributes.

    Allows convenient access to attribute properties for a list of one or more attributes.

    For example, if we have only one attribute we want to access the ``size``, we want to be able
    to do ``attribute_list.size`` instead of ``attribute_list[0].size``.

    Additionally, we can also provide functionality here that we want to perform on a group of
    attributes, like ``open()`` and ``size()``.
    """

    def __getattr__(self, attr: str) -> Any:
        if len(self) == 0:
            raise AttributeError("Attribute not found")

        if hasattr(self[0], attr):
            return getattr(self[0], attr)

        return super().__getattribute__(attr)

    def open(self, allocated: bool = False) -> BinaryIO:
        """Open the data streams on a list of attributes, resident or non-resident.

        Args:
            allocated: Use the actual stream size or the allocated stream size (i.e. include slack space or not).

        Returns:
            A file-like object for the data of this list of attributes.
        """
        if self.header.resident:
            return self.header.open()

        attrs = self._get_stream_attrs()
        # len(attrs) is always >= 1, because if we were empty we would've errored already, and if we have resident
        # attributes, we wouldn't reach this path
        ntfs = attrs[0].record.ntfs

        ensure_volume(ntfs)

        runs = self._get_dataruns(attrs)
        size = attrs[0].header.allocated_size if allocated else attrs[0].header.size

        if attrs[0].header.flags & ATTRIBUTE_FLAG_COMPRESSION_MASK:
            return CompressedRunlistStream(
                ntfs.fh,
                runs,
                size,
                ntfs.cluster_size,
                attrs[0].header.compression_unit,
            )

        return RunlistStream(
            ntfs.fh,
            runs,
            size,
            ntfs.cluster_size,
        )

    def size(self, allocated: bool = False) -> int:
        """Retrieve the data stream size for this list of attributes.

        Args:
            allocated: Return the actual stream size or the allocated stream size (i.e. include slack space or not).

        Returns:
            The requested stream size.
        """
        if self.header.resident:
            return self.header.size

        attrs = self._get_stream_attrs()
        return attrs[0].header.allocated_size if allocated else attrs[0].header.size

    def dataruns(self) -> list[tuple[int, int]]:
        """Get the dataruns for this list of attributes.

        Raises:
            TypeError: If attribute is resident.
        """
        if self.header.resident:
            raise TypeError("Attribute is resident and has no dataruns")

        return self._get_dataruns()

    def _get_stream_attrs(self) -> list[Attribute]:
        return sorted((attr for attr in self if not attr.header.resident), key=lambda attr: attr.header.lowest_vcn)

    def _get_dataruns(self, attrs: list[Attribute] | None = None) -> list[tuple[int, int]]:
        attrs = attrs or self._get_stream_attrs()

        runs = []
        for attr in self._get_stream_attrs():
            runs += attr.header.dataruns()

        return runs


def apply_fixup(data: bytes) -> bytes:
    """Parse and apply fixup data from ``MULTI_SECTOR_HEADER`` to the given bytes.

    Args:
        data: The bytes to fixup

    Returns:
        The fixed up bytes.
    """
    data = bytearray(data)
    view = memoryview(data)

    fixup_offset = struct.unpack("<H", view[4:6])[0]
    fixup_count = len(data) >> SECTOR_SHIFT

    if (
        # The fixup offset must be even
        fixup_offset & 1
        # The end of the fixup array should not exceed SECTOR_SIZE
        or fixup_offset + (fixup_count + 1) * 2 > SECTOR_SIZE
        # Must have at least one fixup
        or fixup_count == 0
        # And the amount of fixups must match the amount of sectors we have data of
        or fixup_count * SECTOR_SIZE > len(data)
    ):
        raise ValueError(f"Fixup offset out of range: {fixup_offset}")

    fixup = view[fixup_offset:]
    sample = fixup[:2]
    fixup = fixup[2:]

    ptr = view[SECTOR_SIZE - 2 :]
    for _ in range(fixup_count):
        if ptr[:2] != sample:
            raise ValueError("Fixup mismatch")

        ptr[:2] = fixup[:2]
        fixup = fixup[2:]
        ptr = ptr[SECTOR_SIZE:]

    return bytes(data)


def ensure_volume(ntfs: NTFS) -> None:
    """Check if a volume is available for reading.

    A volume in this context refers to a disk or other file that contains the raw NTFS data, not contained
    in system files like the ``$MFT``.

    Raises:
        VolumeNotAvailableError: If a volume is not available.
    """
    if not ntfs or not ntfs.fh:
        raise VolumeNotAvailableError


def get_full_path(mft: Mft, name: str, parent: c_ntfs._MFT_SEGMENT_REFERENCE, seen: set[str] | None = None) -> str:
    """Walk up parent file references to construct a full path.

    Args:
        mft: The MFT object to use for looking up file references.
        name: The file name to use.
        parent: The parent reference to start backtracking from.

    Raises:
        FilenameNotAvailableError: If an MFT record has no filename.
    """
    seen = seen or set()

    path = [name]

    while True:
        parent_ref = segment_reference(parent)
        if parent_ref == FILE_NUMBER_ROOT:
            break

        if parent_ref in seen:
            path.append("<recursion>")
            break

        seen.add(parent_ref)

        try:
            record = mft.get(parent_ref)
            if not record.filename:
                raise FilenameNotAvailableError("No filename")  # noqa: TRY301

            if record.header.SequenceNumber != parent.SequenceNumber:
                path.append(f"<broken_reference_0x{parent_ref:x}#{parent.SequenceNumber}>")
                break

            path.append(record.filename)

            parent = record.attributes.FILE_NAME.attr.ParentDirectory
        except Exception:
            path.append(f"<unknown_segment_0x{parent_ref:x}>")
            break

    return "\\".join(path[::-1])


def ts_to_ns(ts: int) -> int:
    """Convert Windows timestamps to nanosecond timestamps."""
    return (ts * 100) - 11644473600000000000


def segment_reference(reference: c_ntfs._MFT_SEGMENT_REFERENCE) -> int:
    """Helper to calculate the complete segment number from a cstruct MFT segment reference.

    Args:
        reference: A cstruct _MFT_SEGMENT_REFERENCE instance to return the complete segment number of.
    """
    return reference.SegmentNumberLowPart | (reference.SegmentNumberHighPart << 32)


def varint(buf: bytes) -> int:
    """Parse variable integers.

    Dataruns in NTFS are stored as a tuple of variable sized integers. The size of each integer is
    stored in the first byte, 4 bits for each integer. This logic can be seen in
    :func:`AttributeHeader.dataruns <dissect.ntfs.attr.AttributeHeader.dataruns>`.

    This function only parses those variable amount of bytes into actual integers. To do that, we
    simply pad the bytes to 8 bytes long and parse it as a signed 64 bit integer. We pad with 0xff
    if the number is negative and 0x00 otherwise.

    Args:
        buf: The byte buffer to parse a varint from.
    """
    if len(buf) < 8:
        buf += (b"\xff" if buf[-1] & 0x80 else b"\x00") * (8 - len(buf))

    return struct.unpack("<q", buf)[0]


def bsf(value: int) -> int:
    """Count the number of trailing zero bits in an integer of a given size.

    Args:
        value: The integer to count trailing zero bits in.
    """
    return (value & -value).bit_length() - 1 if value else 0



from __future__ import annotations

import io
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.util.stream import RangeStream, RunlistStream
from dissect.util.ts import wintimestamp

from dissect.ntfs.c_ntfs import ATTRIBUTE_TYPE_CODE, IO_REPARSE_TAG, c_ntfs
from dissect.ntfs.exceptions import MftNotAvailableError, VolumeNotAvailableError
from dissect.ntfs.util import ensure_volume, get_full_path, segment_reference, ts_to_ns, varint

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from dissect.ntfs.mft import MftRecord


class Attribute:
    """Parse and interact with MFT attributes.

    Wrapper for an :class:`AttributeHeader` and :class:`AttributeRecord` combination.

    Args:
        record: The MFT record this attribute belongs to.
        header: The AttributeHeader for this Attribute.
    """

    __slots__ = ("attribute", "header", "record")

    def __init__(self, header: AttributeHeader, record: MftRecord | None = None):
        self.header = header
        self.record = record
        self.attribute = None

        if header.type in ATTRIBUTE_CLASS_MAP:
            # The attribute may be non-resident when we have no volume
            try:
                self.attribute = AttributeRecord.from_fh(header.open(), header.type, record)
            except VolumeNotAvailableError:
                pass

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.attribute, attr)

    def __repr__(self) -> str:
        return f"<${self.header.type.name} name={self.header.name}>"

    @classmethod
    def from_fh(cls, fh: BinaryIO, record: MftRecord | None = None) -> Attribute:
        """Parse an attribute from a file-like object.

        Args:
            fh: The file-like object to parse an attribute from.
            record: The MFT record this attribute belongs to.
        """
        return cls(AttributeHeader(fh, 0, record), record)

    @classmethod
    def from_bytes(cls, data: bytes, record: MftRecord | None = None) -> Attribute:
        """Parse an attribute from bytes.

        Args:
            data: The bytes to parse.
            record: The MFT record this attribute belongs to.
        """
        return Attribute.from_fh(io.BytesIO(data), record)

    @property
    def type(self) -> ATTRIBUTE_TYPE_CODE:
        """Return the attribute type."""
        return self.header.type

    @property
    def resident(self) -> bool:
        """Return whether this attribute is resident or not."""
        return self.header.resident

    @property
    def name(self) -> str:
        """Return the name of this attribute."""
        return self.header.name

    def dataruns(self) -> list[tuple[int, int]]:
        """Return the dataruns of this attribute, if non-resident.

        Raises:
            TypeError: If attribute is resident.
        """
        return self.header.dataruns()

    def open(self) -> BinaryIO:
        """Open a file-like object for this attribute's data.

        Raises:
            VolumeNotAvailableError: If no volume is available.
        """
        return self.header.open()

    def data(self) -> bytes:
        """Read and return all the data of this attribute.

        Raises:
            VolumeNotAvailableError: If no volume is available.
        """
        return self.header.data()


class AttributeHeader:
    """Parse attribute headers.

    Args:
        record: The MFT record this attribute belongs to.
        fh: The file-like object to parse an attribute header from.
        offset: The offset in the file-like object to parse an attribute header from.
    """

    __slots__ = ("fh", "header", "offset", "record")

    def __init__(self, fh: BinaryIO, offset: int, record: MftRecord | None = None):
        self.fh = fh
        self.offset = offset
        self.record = record

        fh.seek(offset)
        self.header = c_ntfs._ATTRIBUTE_RECORD_HEADER(fh)

    def __repr__(self) -> str:
        return f"<${self.type.name} size={self.size}>"

    @classmethod
    def from_bytes(cls, data: bytes, record: MftRecord | None = None) -> AttributeHeader:
        """Parse an attribute header from bytes.

        Args:
            data: The bytes to parse.
            record: The MFT record this attribute belongs to.
        """
        return cls(io.BytesIO(data), 0, record)

    @property
    def type(self) -> ATTRIBUTE_TYPE_CODE:
        """Return the attribute type."""
        return self.header.TypeCode

    @property
    def resident(self) -> bool:
        """Return whether this attribute is resident or not."""
        return self.header.FormCode == 0

    @property
    def record_length(self) -> int:
        """Return the record length of this attribute."""
        return self.header.RecordLength

    @property
    def name(self) -> str:
        """Return the name of this attribute."""
        self.fh.seek(self.offset + self.header.NameOffset)
        return self.fh.read(self.header.NameLength * 2).decode("utf-16-le")

    @property
    def flags(self) -> int:
        """Return the attribute flags."""
        return self.header.Flags

    @property
    def size(self) -> int:
        """Return the data size of this attribute."""
        return self.header.Form.Resident.ValueLength if self.resident else self.header.Form.Nonresident.FileSize

    @property
    def allocated_size(self) -> int | None:
        """Return the allocated size if non-resident, else None."""
        return self.header.Form.Nonresident.AllocatedLength if not self.resident else None

    @property
    def lowest_vcn(self) -> int | None:
        """Return the lowest VCN if non-resident, else None."""
        return self.header.Form.Nonresident.LowestVcn if not self.resident else None

    @property
    def highest_vcn(self) -> int | None:
        """Return the highest VCN if non-resident, else None."""
        return self.header.Form.Nonresident.HighestVcn if not self.resident else None

    @property
    def compression_unit(self) -> int | None:
        """Return the compression unit if non-resident, else None."""
        return self.header.Form.Nonresident.CompressionUnit if not self.resident else None

    def dataruns(self) -> list[tuple[int, int]]:
        """Return the dataruns of this attribute, if non-resident.

        Raises:
            TypeError: If attribute is resident.
        """
        if self.resident:
            raise TypeError("Resident attributes don't have dataruns")

        fh = self.fh

        fh.seek(self.offset + self.header.Form.Nonresident.MappingPairsOffset)

        runs = []
        run_offset = None
        run_size = None
        offset = 0

        while True:
            value = fh.read(1)[0]  # Get the integer value
            if value == 0:
                break

            size_len = value & 0xF
            offset_len = (value >> 4) & 0xF

            run_size = varint(fh.read(size_len))
            if offset_len == 0:  # Sparse run
                run_offset = None
            else:
                run_offset = offset = offset + varint(fh.read(offset_len))

            runs.append((run_offset, run_size))

        return runs

    def open(self) -> BinaryIO:
        """Open a file-like object for this attribute's data.

        Raises:
            VolumeNotAvailableError: If no volume is available.
        """
        if self.resident:
            return RangeStream(
                self.fh,
                self.offset + self.header.Form.Resident.ValueOffset,
                self.size,
            )

        ntfs = self.record.ntfs if self.record else None
        ensure_volume(ntfs)

        return RunlistStream(
            ntfs.fh,
            self.dataruns(),
            self.size,
            ntfs.cluster_size,
        )

    def data(self) -> bytes:
        """Read and return all the data of this attribute.

        Raises:
            VolumeNotAvailableError: If no volume is available.
        """
        return self.open().read()


class AttributeRecord:
    """Parse attribute records.

    Args:
        record: The MFT record this attribute belongs to.
        fh: The file-like object to parse an attribute record from.
    """

    __slots__ = ("record",)

    def __init__(self, fh: BinaryIO, record: MftRecord | None = None):
        self.record = record

    @classmethod
    def from_fh(cls, fh: BinaryIO, attr_type: ATTRIBUTE_TYPE_CODE, record: MftRecord | None = None) -> AttributeRecord:
        """Parse an attribute from a file-like object.

        Selects a more specific :class:`AttributeRecord` class if one is available for the given attribute type.

        Args:
            fh: The file-like object to parse an attribute from.
            attr_type: The attribute type to parse.
            record: The MFT record this attribute belongs to.
        """
        return ATTRIBUTE_CLASS_MAP.get(attr_type, cls)(fh, record)


class AttributeList(AttributeRecord):
    """Specific :class:`AttributeRecord` parser for ``$ATTRIBUTE_LIST``."""

    __slots__ = ("entries",)

    def __init__(self, fh: BinaryIO, record: MftRecord | None = None):
        super().__init__(fh, record)

        offset = 0
        self.entries = []

        while True:
            fh.seek(offset)

            try:
                entry = c_ntfs._ATTRIBUTE_LIST_ENTRY(fh)
            except EOFError:
                break

            if entry.RecordLength == 0:
                break

            self.entries.append(entry)
            offset += entry.RecordLength

    def __repr__(self) -> str:
        return "<$ATTRIBUTE_LIST>"

    def attributes(self) -> Iterator[Attribute]:
        """Iterate all entries within this ``$ATTRIBUTE_LIST`` and yield all embedded attributes."""
        if not self.record:
            raise MftNotAvailableError("Can't iterate $ATTRIBUTE_LIST attributes without a bounded MFT record")

        seen = {self.record.segment}

        for entry in self.entries:
            segment = segment_reference(entry.SegmentReference)
            if segment == 0 or segment in seen:
                continue

            seen.add(segment)
            record = self.record.ntfs.mft(segment)

            for attr_list in record.attributes.values():
                yield from attr_list


class StandardInformation(AttributeRecord):
    """Specific :class:`AttributeRecord` parser for ``$STANDARD_INFORMATION``."""

    __slots__ = ("attr",)

    def __init__(self, fh: BinaryIO, record: MftRecord | None = None):
        super().__init__(fh, record)
        # Data can be less than the _STANDARD_INFORMATION structure size, so pad remaining fields with null bytes
        data = fh.read().ljust(len(c_ntfs._STANDARD_INFORMATION), b"\x00")
        self.attr = c_ntfs._STANDARD_INFORMATION(data)

    def __repr__(self) -> str:
        return "<$STANDARD_INFORMATION>"

    @property
    def creation_time(self) -> datetime:
        """Return the ``$STANDARD_INFORMATION`` ``CreationTime``."""
        return wintimestamp(self.attr.CreationTime)

    @property
    def creation_time_ns(self) -> int:
        """Return the ``$STANDARD_INFORMATION`` ``CreationTime`` in nanoseconds."""
        return ts_to_ns(self.attr.CreationTime)

    @property
    def last_modification_time(self) -> datetime:
        """Return the ``$STANDARD_INFORMATION`` ``LastModificationTime``."""
        return wintimestamp(self.attr.LastModificationTime)

    @property
    def last_modification_time_ns(self) -> int:
        """Return the ``$STANDARD_INFORMATION`` ``LastModificationTime`` in nanoseconds."""
        return ts_to_ns(self.attr.LastModificationTime)

    @property
    def last_change_time(self) -> datetime:
        """Return the ``$STANDARD_INFORMATION`` ``LastChangeTime``."""
        return wintimestamp(self.attr.LastChangeTime)

    @property
    def last_change_time_ns(self) -> int:
        """Return the ``$STANDARD_INFORMATION`` ``LastChangeTime`` in nanoseconds."""
        return ts_to_ns(self.attr.LastChangeTime)

    @property
    def last_access_time(self) -> datetime:
        """Return the ``$STANDARD_INFORMATION`` ``LastAccessTime``."""
        return wintimestamp(self.attr.LastAccessTime)

    @property
    def last_access_time_ns(self) -> int:
        """Return the ``$STANDARD_INFORMATION`` ``LastAccessTime`` in nanoseconds."""
        return ts_to_ns(self.attr.LastAccessTime)

    @property
    def file_attributes(self) -> int:
        """Return the ``$STANDARD_INFORMATION`` ``FileAttributes``."""
        return self.attr.FileAttributes

    @property
    def owner_id(self) -> int:
        """Return the ``$STANDARD_INFORMATION`` ``OwnerId``."""
        return self.attr.OwnerId

    @property
    def security_id(self) -> int:
        """Return the ``$STANDARD_INFORMATION`` ``SecurityId``."""
        return self.attr.SecurityId


class FileName(AttributeRecord):
    """Specific :class:`AttributeRecord` parser for ``$FILE_NAME``."""

    __slots__ = ("attr",)

    def __init__(self, fh: BinaryIO, record: MftRecord | None = None):
        super().__init__(fh, record)
        data = fh.read().ljust(len(c_ntfs.STANDARD_INFORMATION_EX), b"\x00")
        self.attr = c_ntfs._FILE_NAME(data)

    def __repr__(self) -> str:
        return f"<$FILE_NAME {self.file_name}>"

    @property
    def creation_time(self) -> datetime:
        """Return the ``$FILE_NAME``file ``CreationTime``."""
        return wintimestamp(self.attr.CreationTime)

    @property
    def creation_time_ns(self) -> int:
        """Return the ``$FILE_NAME`` file ``CreationTime`` in nanoseconds."""
        return ts_to_ns(self.attr.CreationTime)

    @property
    def last_modification_time(self) -> datetime:
        """Return the ``$FILE_NAME`` file ``LastModificationTime``."""
        return wintimestamp(self.attr.LastModificationTime)

    @property
    def last_modification_time_ns(self) -> int:
        """Return the ``$FILE_NAME`` file ``LastModificationTime`` in nanoseconds."""
        return ts_to_ns(self.attr.LastModificationTime)

    @property
    def last_change_time(self) -> datetime:
        """Return the ``$FILE_NAME`` file ``LastChangeTime``."""
        return wintimestamp(self.attr.LastChangeTime)

    @property
    def last_change_time_ns(self) -> int:
        """Return the ``$FILE_NAME`` file ``LastChangeTime`` in nanoseconds."""
        return ts_to_ns(self.attr.LastChangeTime)

    @property
    def last_access_time(self) -> datetime:
        """Return the ``$FILE_NAME`` file ``LastAccessTime``."""
        return wintimestamp(self.attr.LastAccessTime)

    @property
    def last_access_time_ns(self) -> int:
        """Return the ``$FILE_NAME`` file ``LastAccessTime`` in nanoseconds."""
        return ts_to_ns(self.attr.LastAccessTime)

    @property
    def file_size(self) -> int:
        """Return the ``$FILE_NAME`` file ``FileSize``."""
        return self.attr.FileSize

    @property
    def file_attributes(self) -> int:
        """Return the ``$FILE_NAME`` file ``FileAttributes``."""
        return self.attr.FileAttributes

    @property
    def flags(self) -> int:
        """Return the ``$FILE_NAME`` flags, which can be either ``FILE_NAME_NTFS`` or ``FILE_NAME_DOS``."""
        return self.attr.Flags

    @property
    def file_name(self) -> str:
        """Return the file name string stored in this ``$FILE_NAME`` attribute."""
        return self.attr.FileName

    def full_path(self) -> str:
        """Use the parent directory reference to try to generate a full path from this file name."""
        return get_full_path(self.record.ntfs.mft, self.file_name, self.attr.ParentDirectory)


class ReparsePoint(AttributeRecord):
    """Specific :class:`AttributeRecord` parser for ``$REPARSE_POINT``."""

    __slots__ = ("attr", "buffer", "tag_header")

    def __init__(self, fh: BinaryIO, record: MftRecord | None = None):
        super().__init__(fh, record)
        self.attr = c_ntfs._REPARSE_DATA_BUFFER(fh)
        data = io.BytesIO(fh.read(self.attr.ReparseDataLength))

        self.tag_header = None
        if self.tag == IO_REPARSE_TAG.SYMLINK:
            self.tag_header = c_ntfs._SYMBOLIC_LINK_REPARSE_BUFFER(data)
        elif self.tag == IO_REPARSE_TAG.MOUNT_POINT:
            self.tag_header = c_ntfs._MOUNT_POINT_REPARSE_BUFFER(data)

        self.buffer = data.read()

    def __repr__(self) -> str:
        return f"<$FILE_NAME {self.print_name}>"

    @property
    def tag(self) -> IO_REPARSE_TAG:
        return self.attr.ReparseTag

    @property
    def substitute_name(self) -> str | None:
        if not self.tag_header:
            return None

        offset = self.tag_header.SubstituteNameOffset
        length = self.tag_header.SubstituteNameLength
        return self.buffer[offset : offset + length].decode("utf-16-le")

    @property
    def print_name(self) -> str | None:
        if not self.tag_header:
            return None

        offset = self.tag_header.PrintNameOffset
        length = self.tag_header.PrintNameLength
        return self.buffer[offset : offset + length].decode("utf-16-le")

    @property
    def absolute(self) -> bool:
        if self.tag != IO_REPARSE_TAG.SYMLINK:
            return True

        return self.tag_header.Flags == c_ntfs.SYMLINK_FLAG.ABSOLUTE

    @property
    def relative(self) -> bool:
        if self.tag != IO_REPARSE_TAG.SYMLINK:
            return False

        return self.tag_header.Flags == c_ntfs.SYMLINK_FLAG.RELATIVE


ATTRIBUTE_CLASS_MAP = {
    ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION: StandardInformation,
    ATTRIBUTE_TYPE_CODE.ATTRIBUTE_LIST: AttributeList,
    ATTRIBUTE_TYPE_CODE.FILE_NAME: FileName,
    ATTRIBUTE_TYPE_CODE.REPARSE_POINT: ReparsePoint,
}



from __future__ import annotations

import io
from enum import Enum, auto
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.ntfs.attr import AttributeRecord
from dissect.ntfs.c_ntfs import ATTRIBUTE_TYPE_CODE, COLLATION, INDEX_ENTRY_END, INDEX_ENTRY_NODE, SECTOR_SHIFT, c_ntfs
from dissect.ntfs.exceptions import (
    BrokenIndexError,
    Error,
    FileNotFoundError,
    MftNotAvailableError,
)
from dissect.ntfs.util import apply_fixup, segment_reference

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from dissect.ntfs.mft import MftRecord


class Match(Enum):
    Less = auto()
    Equal = auto()
    Greater = auto()


class Index:
    """Open an index with he given name on the given MFT record.

    Args:
        name: The index to open.

    Raises:
        FileNotFoundError: If no index with that name can be found.
    """

    def __init__(self, record: MftRecord, name: str):
        self.record = record
        self.name = name

        self.root = IndexRoot(self, self.record.open(name, ATTRIBUTE_TYPE_CODE.INDEX_ROOT))

        if self.record.ntfs and self.record.ntfs.cluster_size <= self.root.bytes_per_index_buffer:
            self._vcn_size_shift = self.record.ntfs.cluster_size_shift
        else:
            self._vcn_size_shift = SECTOR_SHIFT

        try:
            self._index_stream = self.record.open(self.name, ATTRIBUTE_TYPE_CODE.INDEX_ALLOCATION)
        except FileNotFoundError:
            self._index_stream = None

        self.index_buffer = lru_cache(128)(self.index_buffer)

    def __iter__(self) -> Iterator[IndexEntry]:
        return self.entries()

    def index_buffer(self, vcn: int) -> IndexBuffer:
        """Return the :class:`IndexBuffer` at the specified cluster number.

        Args:
            vcn: The virtual cluster number within the index allocation to read.

        Raises:
            FileNotFoundError: If this index has no index allocation.
        """
        if not self._index_stream:
            raise FileNotFoundError(f"Index has no index allocation: {self.name}")

        return IndexBuffer(self, self._index_stream, vcn << self._vcn_size_shift, self.root.bytes_per_index_buffer)

    def search(
        self, value: Any, exact: bool = True, cmp: Callable[[IndexEntry, Any], Match] | None = None
    ) -> IndexEntry:
        """Perform a binary search on this index.

        Returns the matching node if performing an exact search. Otherwise return the first match that is greater
        than the search value.

        Args:
            value: The key to search.
            exact: Result must be an exact match.
            cmp: Optional custom comparator function.

        Raises:
            NotImplementedError: If there is no collation (comparator) function for the collation rule of this index.
            KeyError: If an exact match was requested but not found.
        """
        cmp_functions = {
            COLLATION.FILE_NAME: _cmp_filename,
            COLLATION.NTOFS_ULONG: _cmp_ulong,
        }

        cmp = cmp or cmp_functions.get(self.root.collation_rule)
        if not cmp:
            raise NotImplementedError(f"No collation function for collation rule: {self.root.collation_rule}")

        search_value = value
        if cmp == _cmp_filename:
            search_value = value.upper()

        entries = list(self.root.entries())

        while True:
            entry = _bsearch(entries, search_value, cmp)
            if not entry.is_node or (not entry.is_end and cmp(entry, search_value) == Match.Equal):
                break
            entries = list(self.index_buffer(entry.node_vcn).entries())

        if exact and (entry.is_end or cmp(entry, search_value) != Match.Equal):
            raise KeyError(f"Value not found: {value}")

        return entry

    def entries(self) -> Iterator[IndexEntry]:
        """Yield all :class:`IndexEntry`'s in this :class:`Index`."""

        for entry in self.root.entries():
            if entry.is_end:
                break

            yield entry

        if self._index_stream:
            vcn = 0
            while True:
                try:
                    for entry in self.index_buffer(vcn).entries():
                        if entry.is_end:
                            break

                        yield entry
                except Error:
                    pass
                except EOFError:
                    break

                vcn += 1


class IndexRoot:
    """Represents the ``$INDEX_ROOT``.

    Args:
        index: The :class:`Index`` class instance this :class:`IndexRoot` belongs to.
        fh: The file-like object to parse an index root on.
    """

    def __init__(self, index: Index, fh: BinaryIO):
        self.index = index
        self.fh = fh

        self.header = c_ntfs._INDEX_ROOT(fh)

    @property
    def attribute_type(self) -> ATTRIBUTE_TYPE_CODE:
        """Return the indexed attribute type."""
        return self.header.AttributeType

    @property
    def collation_rule(self) -> COLLATION:
        """Return the collation rule."""
        return self.header.CollationRule

    @property
    def bytes_per_index_buffer(self) -> int:
        """Return the size of an index buffer in the index allocation in bytes."""
        return self.header.BytesPerIndexBuffer

    @property
    def clusters_per_index_buffer(self) -> int:
        """Return the size of an index buffer in the index allocation in clusters."""
        return self.header.ClustersPerIndexBuffer

    def entries(self) -> Iterator[IndexEntry]:
        """Yield all :class:`IndexEntry`'s in this :class:`IndexRoot`."""
        yield from _iter_entries(
            self.index,
            self.fh,
            # Offset starts from the _INDEX_HEADER
            len(c_ntfs._INDEX_ROOT),
            self.header.IndexHeader.TotalSizeOfEntries,
        )


class IndexBuffer:
    """Represent an index buffer in ``$INDEX_ALLOCATION``.

    Args:
        index: The :class:`Index` class instance this :class:`IndexRoot` belongs to.
        fh: The file-like object of ``$INDEX_ALLOCATION``.
        offset: The offset in bytes to the index buffer on the file-like object we want to read.
        size: The size of the index buffer in bytes.

    Raises:
        EOFError: If there's not enough data available to read an index buffer.
        BrokenIndexError: If the index buffer doesn't start with the expected magic value.
    """

    def __init__(self, index: Index, fh: BinaryIO, offset: int, size: int):
        self.index = index
        self.offset = offset
        self.size = size

        fh.seek(offset)
        buf = fh.read(size)

        if len(buf) != size:
            raise EOFError

        if buf[:4] != b"INDX":
            raise BrokenIndexError("Broken INDX header")

        self.data = apply_fixup(buf)
        self.header = c_ntfs._INDEX_ALLOCATION_BUFFER(self.data)

    def entries(self) -> Iterator[IndexEntry]:
        """Yield all :class:`IndexEntry`'s in this :class:`IndexBuffer`."""
        yield from _iter_entries(
            self.index,
            io.BytesIO(self.data),
            # Offset starts from the _INDEX_HEADER
            self.header.IndexHeader.FirstEntryOffset + 0x18,
            self.header.IndexHeader.TotalSizeOfEntries,
        )


class IndexEntry:
    """Parse and interact with index entries.

    Args:
        index: The :class:`Index` class instance this :class:`IndexEntry` belongs to.
        fh: The file-like object to parse an index entry on.
        offset: The offset in the file-like object to parse an index entry at.
    """

    def __init__(self, index: Index, fh: BinaryIO, offset: int):
        self.index = index
        self.fh = fh
        self.offset = offset

        fh.seek(offset)
        self.header = c_ntfs._INDEX_ENTRY(fh)
        self.buf = fh.read(self.header.Length - len(c_ntfs._INDEX_ENTRY))

    def dereference(self) -> MftRecord:
        """Dereference this :class:`IndexEntry` to the MFT record it points to.

        Note that the file reference is a union with the data part so only access this if you know the entry has
        a file reference and not a data part.

        Raises:
            MftNotAvailableError: If no MFT is available.
        """
        record = self.index.record
        if not record or not record.ntfs or not record.ntfs.mft:
            raise MftNotAvailableError

        return record.ntfs.mft.get(segment_reference(self.header.FileReference))

    @cached_property
    def key(self) -> bytes:
        """Return the index key of this entry."""
        return self.buf[: self.header.KeyLength]

    @cached_property
    def data(self) -> bytes:
        """Return the data part of this entry.

        Note that the data part is a union with the file reference, so only access this if you know the entry has
        data and not a file reference.
        """
        offset = self.header.DataOffset - len(c_ntfs._INDEX_ENTRY)
        return self.buf[offset : offset + self.header.DataLength]

    @cached_property
    def attribute(self) -> AttributeRecord | None:
        """Return the :class:`dissect.ntfs.attr.AttributeRecord` of the attribute contained in this entry."""
        if self.key_length and self.index.root.attribute_type:
            return AttributeRecord.from_fh(
                io.BytesIO(self.buf),
                self.index.root.attribute_type,
                record=self.index.record,
            )
        return None

    @property
    def is_end(self) -> bool:
        """Return whether this entry marks the end."""
        return bool(self.header.Flags & INDEX_ENTRY_END) or self.header.Length == 0

    @property
    def is_node(self) -> bool:
        """Return whether this entry is a node."""
        return bool(self.header.Flags & INDEX_ENTRY_NODE)

    @property
    def node_vcn(self) -> int:
        """Return the node VCN if this entry is a node."""
        if self.is_node:
            return c_ntfs.ULONG64(memoryview(self.buf)[-8:])
        return None

    @property
    def length(self) -> int:
        """Return the length of this index entry."""
        return self.header.Length

    @property
    def key_length(self) -> int:
        """Return the length of this index entry."""
        return self.header.KeyLength


def _iter_entries(index: Index, fh: BinaryIO, offset: int, size: int) -> Iterator[IndexEntry]:
    max_offset = offset + size
    while offset < max_offset:
        try:
            entry = IndexEntry(index, fh, offset)
        except EOFError:
            break

        yield entry

        if entry.is_end:
            break

        offset += entry.length


def _bsearch(entries: list[IndexEntry], value: Any, cmp: Callable[[IndexEntry, Any], Match]) -> IndexEntry:
    min_idx = 0
    max_idx = len(entries) - 1

    while min_idx != max_idx:
        test_idx = min_idx + (max_idx - min_idx) // 2
        test_entry = entries[test_idx]

        result = cmp(test_entry, value)
        if not test_entry.is_end and result == Match.Greater:
            min_idx = test_idx + 1
        elif result == Match.Equal:
            return test_entry
        else:
            max_idx = test_idx

    return entries[min_idx]


def _cmp_filename(entry: IndexEntry, value: str) -> Match:
    # We could parse an entire _FILE_NAME here but for performance reasons we cheat
    # FileNameLength is at offset 64 and is a single byte
    length = entry.buf[64]
    # FileName starts at offset 66 and is double the FileNameLength because of UTF-16 encoding
    test_value = entry.buf[66 : 66 + (length * 2)].decode("utf-16-le").upper()

    if value < test_value:
        return Match.Less
    if value == test_value:
        return Match.Equal
    return Match.Greater


def _cmp_ulong(entry: IndexEntry, value: int) -> Match:
    key = entry.key
    if len(key) != 4:
        raise ValueError(f"Invalid key length for ULONG collation: {len(key)}")
    test_value = int.from_bytes(key, "little")

    if value < test_value:
        return Match.Less
    if value == test_value:
        return Match.Equal
    return Match.Greater



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.util.stream import RunlistStream
from dissect.util.ts import wintimestamp

from dissect.ntfs.c_ntfs import USN_PAGE_SIZE, c_ntfs
from dissect.ntfs.exceptions import Error
from dissect.ntfs.util import segment_reference, ts_to_ns

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from dissect.ntfs.mft import MftRecord
    from dissect.ntfs.ntfs import NTFS


class UsnJrnl:
    """Parse the USN journal from a file-like object of the ``$UsnJrnl:$J`` stream.

    Args:
        fh: A file-like object of the $UsnJrnl:$J stream.
        ntfs: An optional :class:`~dissect.ntfs.ntfs.NTFS` class instance, used for resolving file paths.
    """

    def __init__(self, fh: BinaryIO, ntfs: NTFS | None = None):
        self.fh = fh
        self.ntfs = ntfs

    def records(self) -> Iterator[UsnRecord]:
        """Yield all parsed USN records.

        Only yields version 2 USN records, other record versions are ignored.
        """
        fh = self.fh
        offset = 0

        if isinstance(fh, RunlistStream):
            for run_offset, run_size in fh.runlist:
                if run_offset is not None:
                    break

                offset += run_size * fh.block_size

        while True:
            fh.seek(offset)

            if fh.read(4) == b"\x00" * 4:
                # Increment to USN_PAGE_SIZE
                offset += USN_PAGE_SIZE - (offset % USN_PAGE_SIZE)
                continue

            try:
                record = UsnRecord(self, fh, offset)
                if record.header.MajorVersion == 2:
                    yield record
            except EOFError:
                break

            offset += record.record.RecordLength
            if offset % 8:
                offset += -(offset) & (8 - 1)


class UsnRecord:
    """Parse a USN record from a file-like object and offset.

    Args:
        usnjrnl: The :class:`UsnJrnl` class this record is parsed from.
        fh: The file-like object to parse a USN record from.
        offset: The offset to parse a USN record at.
    """

    def __init__(self, usnjrnl: UsnJrnl, fh: BinaryIO, offset: int):
        self.usnjrnl = usnjrnl
        self.offset = offset
        self.extents = []

        fh.seek(offset)
        self.header = c_ntfs.USN_RECORD_COMMON_HEADER(fh)

        fh.seek(offset)
        if self.header.MajorVersion == 2:
            self.record = c_ntfs.USN_RECORD_V2(fh)
        elif self.header.MajorVersion == 3:
            self.record = c_ntfs.USN_RECORD_V3(fh)
        elif self.header.MajorVersion == 4:
            self.record = c_ntfs.USN_RECORD_V4(fh)
            for _ in range(self.record.NumberOfExtents):
                self.extents.append(c_ntfs.USN_RECORD_EXTENT(fh))
        else:
            raise ValueError(f"Unsupported USN_RECORD version: {self.header.MajorVersion}")

        if self.header.MajorVersion in (2, 3):
            fh.seek(offset + self.record.FileNameOffset)
            self.filename = fh.read(self.record.FileNameLength).decode("utf-16-le")
        else:
            self.filename = None

    def __repr__(self) -> str:
        return f"<UsnRecord {self.record.Usn}>"

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.record, attr)

    @cached_property
    def file(self) -> MftRecord | None:
        if self.usnjrnl.ntfs and self.usnjrnl.ntfs.mft:
            return self.usnjrnl.ntfs.mft(self.record.FileReferenceNumber)
        return None

    @cached_property
    def parent(self) -> MftRecord | None:
        if self.usnjrnl.ntfs and self.usnjrnl.ntfs.mft:
            return self.usnjrnl.ntfs.mft(self.record.ParentFileReferenceNumber)
        return None

    @property
    def timestamp(self) -> datetime:
        return wintimestamp(self.record.TimeStamp)

    @property
    def timestamp_ns(self) -> int:
        return ts_to_ns(self.record.TimeStamp)

    @cached_property
    def full_path(self) -> str:
        try:
            parent = self.parent
        except Error:
            parent = None

        ref = segment_reference(self.record.ParentFileReferenceNumber)
        if parent is None:
            parent_path = f"<unavailable_reference_0x{ref:x}#{self.record.ParentFileReferenceNumber.SequenceNumber}>"
        elif parent.header.SequenceNumber == self.record.ParentFileReferenceNumber.SequenceNumber:
            parent_path = parent.full_path()
        else:
            parent_path = f"<broken_reference_0x{ref:x}#{self.record.ParentFileReferenceNumber.SequenceNumber}>"

        return f"{parent_path}\\{self.filename}"



from __future__ import annotations

import datetime

import pytest

from dissect.ntfs.attr import Attribute, FileName, StandardInformation
from dissect.ntfs.c_ntfs import ATTRIBUTE_TYPE_CODE, IO_REPARSE_TAG
from dissect.ntfs.exceptions import VolumeNotAvailableError


def test_attributes() -> None:
    # Single $STANDARD_INFORMATION attribute
    data = bytes.fromhex(
        "100000006000000000001800000000004800000018000000d2145d665666d801"
        "d2145d665666d801d2145d665666d801d2145d665666d8010600000000000000"
        "0000000000000000000000000001000000000000000000000000000000000000"
    )
    attr = Attribute.from_bytes(data)
    assert attr.type == ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION

    assert isinstance(attr.attribute, StandardInformation)
    assert attr.creation_time == datetime.datetime(2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc)
    assert attr.creation_time_ns == 1652397427802645000
    assert attr.last_modification_time == datetime.datetime(
        2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc
    )
    assert attr.last_modification_time_ns == 1652397427802645000
    assert attr.last_change_time == datetime.datetime(2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc)
    assert attr.last_change_time_ns == 1652397427802645000
    assert attr.last_access_time == datetime.datetime(2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc)
    assert attr.last_access_time_ns == 1652397427802645000

    # Single $FILE_NAME attribute
    data = bytes.fromhex(
        "300000006800000000001800000003004a000000180001000500000000000500"
        "d2145d665666d801d2145d665666d801d2145d665666d801d2145d665666d801"
        "004000000000000000400000000000000600000000000000040324004d004600"
        "5400000000000000"
    )
    attr = Attribute.from_bytes(data)
    assert attr.type == ATTRIBUTE_TYPE_CODE.FILE_NAME

    assert isinstance(attr.attribute, FileName)
    assert attr.file_name == "$MFT"
    assert attr.creation_time == datetime.datetime(2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc)
    assert attr.last_modification_time == datetime.datetime(
        2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc
    )
    assert attr.last_change_time == datetime.datetime(2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc)
    assert attr.last_access_time == datetime.datetime(2022, 5, 12, 23, 17, 7, 802645, tzinfo=datetime.timezone.utc)
    assert attr.file_size == 16384

    # Single $DATA attribute
    data = bytes.fromhex(
        "8000000048000000010040000000060000000000000000003f00000000000000"
        "4000000000000000000004000000000000000400000000000000040000000000"
        "2140550200000000"
    )
    attr = Attribute.from_bytes(data)
    assert attr.type == ATTRIBUTE_TYPE_CODE.DATA

    assert attr.dataruns() == [(597, 64)]

    with pytest.raises(VolumeNotAvailableError):
        attr.open()


def test_reparse_point_moint_point() -> None:
    data = bytes.fromhex(
        "c00000005800000000000000000004004000000018000000030000a038000000"
        "00001a001c0012005c003f003f005c0043003a005c0054006100720067006500"
        "7400000043003a005c00540061007200670065007400"
    )
    attr = Attribute.from_bytes(data)
    assert attr.type == ATTRIBUTE_TYPE_CODE.REPARSE_POINT

    assert attr.tag == IO_REPARSE_TAG.MOUNT_POINT
    assert attr.substitute_name == "\\??\\C:\\Target"
    assert attr.print_name == "C:\\Target"
    assert attr.absolute
    assert not attr.relative


def test_reparse_point_symlink() -> None:
    data = bytes.fromhex(
        "c000000058000000000000000000040040000000180000000c0000a038000000"
        "12001a00000012000000000043003a005c005400610072006700650074005c00"
        "3f003f005c0043003a005c00540061007200670065007400"
    )
    attr = Attribute.from_bytes(data)
    assert attr.type == ATTRIBUTE_TYPE_CODE.REPARSE_POINT

    assert attr.tag == IO_REPARSE_TAG.SYMLINK
    assert attr.substitute_name == "\\??\\C:\\Target"
    assert attr.print_name == "C:\\Target"
    assert attr.absolute
    assert not attr.relative

    data = bytes.fromhex(
        "c00000004800000000000000000004002c000000180000000c0000a024000000"
        "0c000c0000000c00010000005400610072006700650074005400610072006700"
        "6500740000000000"
    )
    attr = Attribute.from_bytes(data)
    assert attr.type == ATTRIBUTE_TYPE_CODE.REPARSE_POINT

    assert attr.tag == IO_REPARSE_TAG.SYMLINK
    assert attr.substitute_name == "Target"
    assert attr.print_name == "Target"
    assert not attr.absolute
    assert attr.relative



from __future__ import annotations

from typing import BinaryIO

import pytest

from dissect.ntfs.ntfs import NTFS
from dissect.ntfs.secure import Secure


def test_secure(ntfs_bin: BinaryIO) -> None:
    fs = NTFS(ntfs_bin)

    assert fs.secure

    sd = fs.secure.lookup(256)
    assert sd.owner == "S-1-5-18"
    assert sd.group == "S-1-5-32-544"

    assert sd.dacl
    assert len(sd.dacl.ace) == 2
    assert sd.dacl.ace[0].type.name == "ACCESS_ALLOWED"
    assert sd.dacl.ace[0].mask == 0x120089
    assert sd.dacl.ace[0].sid == "S-1-5-18"
    assert sd.dacl.ace[1].type.name == "ACCESS_ALLOWED"
    assert sd.dacl.ace[1].mask == 0x120089
    assert sd.dacl.ace[1].sid == "S-1-5-32-544"

    with pytest.raises(KeyError):
        fs.secure.lookup(1337)

    assert len(list(fs.secure.descriptors())) == 24


def test_secure_file(sds_bin: BinaryIO) -> None:
    secure = Secure(sds=sds_bin)

    sd = secure.lookup(256)
    assert sd.owner == "S-1-5-18"
    assert sd.group == "S-1-5-32-544"


def test_secure_complex_acl(sds_complex_bin: BinaryIO) -> None:
    secure = Secure(sds=sds_complex_bin)

    sd = secure.lookup(259)
    assert sd.owner == "S-1-5-21-3090333131-159632407-777084872-1001"
    assert sd.group == "S-1-5-21-3090333131-159632407-777084872-513"
    assert len(sd.sacl.ace) == 2
    assert list(map(repr, sd.sacl.ace)) == [
        "<SYSTEM_MANDATORY_LABEL mask=0x7 sid=S-1-16-4096>",
        "<SYSTEM_ACCESS_FILTER mask=0x1200a9 sid=S-1-1-0>",
    ]
    assert len(sd.dacl.ace) == 3
    assert list(map(repr, sd.dacl.ace)) == [
        "<ACCESS_ALLOWED_COMPOUND mask=0x1f01ff type=COMPOUND_ACE_IMPERSONATION server_sid=S-1-5-11 client_sid=S-1-5-32-545>",  # noqa: E501
        "<ACCESS_ALLOWED_OBJECT mask=0x1f01ff flags=0 object_type=None inherited_object_type=None sid=S-1-1-0>",
        "<ACCESS_ALLOWED_CALLBACK_OBJECT mask=0x1f01ff flags=3 object_type=01234567-89ab-cdef-1111-111111111111 inherited_object_type=22222222-2222-2222-0123-456789abcdef sid=S-1-5-32-545>",  # noqa: E501
    ]


def test_secure_fail() -> None:
    with pytest.raises(ValueError, match="Either record or SDS stream is required"):
        Secure()



from __future__ import annotations

from dissect.ntfs.attr import Attribute
from dissect.ntfs.c_ntfs import ATTRIBUTE_TYPE_CODE
from dissect.ntfs.util import AttributeMap, apply_fixup


def test_fixup() -> None:
    buf = bytearray(
        b"FILE\x30\x00"
        + (b"\x00" * 42)
        + b"\x02\x00\xff\x00\xfe\x00"
        + (b"\x00" * 456)
        + b"\x02\x00"
        + (b"\x00" * 510)
        + b"\x02\x00"
    )
    fixed = apply_fixup(buf)

    assert fixed[510:512] == b"\xff\x00"
    assert fixed[1022:1024] == b"\xfe\x00"

    buf = bytearray(
        b"FILE\x30\x00"
        + (b"\x00" * 42)
        + b"\x02\x00\xff\x00\xfe\x00\xfd\x00\xfc\x00"
        + (b"\x00" * 452)
        + b"\x02\x00"
        + (b"\x00" * 510)
        + b"\x02\x00"
        + (b"\x00" * 510)
        + b"\x02\x00"
        + (b"\x00" * 510)
        + b"\x02\x00"
    )
    fixed = apply_fixup(buf)

    assert fixed[510:512] == b"\xff\x00"
    assert fixed[1022:1024] == b"\xfe\x00"
    assert fixed[1534:1536] == b"\xfd\x00"
    assert fixed[2046:2048] == b"\xfc\x00"


def test_attribute_map() -> None:
    attr_map = AttributeMap()
    assert len(attr_map) == 0
    assert attr_map.STANDARD_INFORMATION == []
    assert attr_map[ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION] == []
    assert ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION not in attr_map

    data = bytes.fromhex(
        "100000006000000000001800000000004800000018000000d2145d665666d801"
        "d2145d665666d801d2145d665666d801d2145d665666d8010600000000000000"
        "0000000000000000000000000001000000000000000000000000000000000000"
    )
    attr = Attribute.from_bytes(data)
    attr_map.add(attr)

    assert len(attr_map) == 1
    assert len(attr_map.STANDARD_INFORMATION) == 1
    assert len(attr_map[ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION]) == 1
    assert attr_map[ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION] == attr_map.STANDARD_INFORMATION
    assert ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION in attr_map

    assert len(attr_map.find("", ATTRIBUTE_TYPE_CODE.STANDARD_INFORMATION)) == 1
